""" Module for creating a GUI window.

Templates for the GUI window can be configured using Qt Designer and should be stored as .ui files in the
./gui_templates directory.

The GUI is designed to hold a number of user (external script) assigned widget attributes (plots, scalars, labels),
and continuously refresh output. A client can access the GUI through the following types of methods:

- Configuration methods: methods to configure widgets. These include assignment of items in the script to specific
    widgets in the GUI, or changing properties of a specific GUI widget. Configuration calls are added to a queue
    and updated directly in the process containing the GUI server, so they are not implemented directly when called
    by the script.
    ^^^ AS A RESULT, SCRIPTS SHOULD ERROR HANDLE WHEN ACCESSING THE GUI IN THE CASE THAT THE CONFIGURATION REQUEST HAS
    NOT BEEN COMPLETED. This can usually be done by checking a KeyError exception, since the gui module uses dicts
    to access configured widgets, but there may be other exceptions depending on the exact implementation
- Data update methods: methods that update the data for various widgets. Again, for reasons described above, these do
    not directly update the output widget but rather an attribute of the specific widget class (Plot, Scalar, etc). The
    thread running the GUI server continuously updates the widget data to the current value of the attribute.
- Activation/deactivation methods: these activate and deactivate the GUI's update to pull from their internal data
    attribute or not. If a GUI widget is active, it continuously sets its output to widget_instance.data (which can be
    externally modified using data update methods). If a GUI widget is inactive, a user can in principle modify data
    from the GUI window.
- Data pull requests: current GUI widget data values are returned to an external script. These should be used to check
    for updates through the GUI window and to change script parameters as a result.

For error handling, all gui update requests should be enclosed in a try/except statement catching the EOFError which
would be thrown in case the GUI crashes. This enables scripts to continue running even if the GUI crashes.
"""

from PyQt5 import QtWidgets, uic, QtCore
import pyqtgraph as pg
import numpy as np
import os
import pickle
import copy
import sys
import socket
import ctypes

# Should help with scaling issues on monitors of differing resolution
if hasattr(QtCore.Qt, 'AA_EnableHighDpiScaling'):
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
if hasattr(QtCore.Qt, 'AA_UseHighDpiPixmaps'):
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


class Window(QtWidgets.QMainWindow):
    """ Main window for GUI.

    This should be instantiated locally to create a GUI window. An app should ALWAYS be instantiated prior to a Window,
    as below:
        app = QtWidgets.QApplication(sys.argv)
        main_window = Window(app, gui_template="my_favorite_gui")
    """

    _gui_directory = "gui_templates"
    _default_template = "count_monitor"

    def __init__(self, app, gui_template=None, run=True):
        """ Instantiates main window object.

        :param app: instance of QApplication class - MUST be instantiated prior to Window
        :param gui_template: (str, optional) name of gui template to use. By default uses self._default_template. Only
            the filename is required (no filepath or extension)
        :param run: (bool, optional) whether or not to run (display) the GUI upon instantiation. Can set to false in
            order to debug and access Window methods directly in an interactive session
        """

        # Initialize parent class QWidgets.QMainWindow
        super(Window, self).__init__()

        self._ui = None  # .ui file to use as a template
        self._app = app  # Application instance onto which to load the GUI. Must be instantiated prior to Window!

        # Holds all widgets assigned to the GUI from an external script
        # Reference is by keyword (widget_label), and the keyword can be used to access the widget once assigned
        self.plots = {}
        self.scalars = {}
        self.labels = {}
        self.event_buttons = {}
        self.containers = {}

        # Configuration queue lists
        # When a script requests to configure a widget (e.g. add or remove a plot, curve, scalar, or label), the request
        # is added to this list and can be implemented by the GUI when it is ready. This prevents overloading the
        # client-server interface
        self._plots_to_assign = []
        self._plots_to_remove = []
        self._curves_to_assign = []
        self._scalars_to_assign = []
        self._labels_to_assign = []
        self._buttons_to_assign = []
        self._containers_to_assign = []
        self._curves_to_remove = []

        # List of Numerical widgets for future use in dynamical step size updating
        self.num_widgets = None

        # Load and run the GUI
        self._load_gui(gui_template=gui_template, run=run)
        self.showNormal()

    def assign_plot(self, plot_widget, plot_label, legend_widget):
        """ Adds plot assignment request to a queue

        :param plot_widget: (str) name of the plot widget to use in the GUI window
        :param plot_label: (str) label to use for future referencing of the plot widget from the self.plots dictionary
        :param legend_widget: (str) name of the pg.GraphicsView instance for use as a legend
        """

        self._plots_to_assign.append((plot_widget, plot_label, legend_widget))

    def clear_plot(self, plot_widget):
        """ Clears all curves and legend items from a plot

        :param plot_widget: (str) name of the plot widget (instance of PYQTgraph) to clear
        """
        self._plots_to_remove.append(plot_widget)

    def assign_curve(self, plot_label, curve_label, error=False):
        """Adds curve assignment to the queue

        :param plot_label: (str) key of plot to assign curve to
        :param curve_label: (str) key to use for curve assignment + legend
        :param error: (bool, optional) whether or not to use error bars
        """

        self._curves_to_assign.append((plot_label, curve_label, error))

    def remove_curve(self, plot_label, curve_label):
        """ Adds a curve removal request to the queue

        :param plot_label: (str) name of plot holding the curve
        :param curve_label: (str) name of the curve to remove
        """

        self._curves_to_remove.append((plot_label, curve_label))

    def assign_scalar(self, scalar_widget, scalar_label):
        """ Adds scalar assignment request to the queue

        :param scalar_widget: (str) name of the scalar object (e.g. QLCDNumber) in the GUI
        :param scalar_label: (str) label for self.numbers dictionary
        """
        self._scalars_to_assign.append((scalar_widget, scalar_label))

    def assign_label(self, label_widget, label_label):
        """ Adds label widget assignment to queue

        :param label_widget: (str) name of label object (e.g. QLabel) in the GUI
        :param label_label: (str) label for self.labels dictionary
        """
        self._labels_to_assign.append((label_widget, label_label))

    def assign_event_button(self, event_widget, event_label):
        """ Adds button assignment request to the queue

        :param event_widget: (str) physical widget name on the .ui file
        :param event_label: (str) keyname to assign to this button for future reference
        """
        self._buttons_to_assign.append((event_widget, event_label))

    def assign_container(self, container_widget, container_label):
        """ Adds Container assignment request to the queue

        Only QListWidget supported so far
        :param container_widget: (str) physical widget name on the .ui file
        :param container_label: (str) keyname to assign to the widget for future reference
        """
        self._containers_to_assign.append((container_widget, container_label))
    
    def set_curve_data(self, data, plot_label, curve_label, error=None):
        """ Sets data to a specific curve (does not update GUI directly)

        :param data: (np.array) 1D or 2D numpy array with data
        :param plot_label: (str) label of the plot
        :param curve_label: (str) curve key of curve to set
        :param error: (np.array, optional) 1D array with error bars, if applicable
        """

        self.plots[plot_label].curves[curve_label].set_curve_data(data, error=error)

    def set_scalar(self, value, scalar_label):
        """ Sets the value of a numerical display internally (does not update)"""
        self.scalars[scalar_label].set_data(value)

    # No decorator here because function returns value
    def get_scalar(self, scalar_label):
        """ Returns the data associated with a scalar

        :return: scalar data
        """

        return self.scalars[scalar_label].get_data()

    def activate_scalar(self, scalar_label):
        """ Tells a scalar to pull data from internal self.data attribute

        :param scalar_label: (str) key for relevant scalar
        """
        self.scalars[scalar_label].activate()

    def deactivate_scalar(self, scalar_label):
        """ Tells a scalar not to update output based on self.data attribute

        :param scalar_label: (str) key for relevant scalar
        """
        self.scalars[scalar_label].deactivate()

    def set_label(self, text, label_label):
        """ Sets a label widgets text

        :param text: (str) Text to set
        :param label_label: (str) Key for the label to set
        """
        self.labels[label_label].set_label(text)

    def get_text(self, label_label):
        """ Returns the text in a textual label widget """

        return self.labels[label_label].get_label()
    
    def was_button_pressed(self, event_label):
        """ Returns whether or not an event button was pressed

        :param event_label: (str) key for button to check
        """

        try:
            return self.event_buttons[event_label].get_state()

        #If the button has not been assigned, just return false
        except KeyError:
            return False

    def change_button_background_color(self, event_label, color):
        """ Change background color of button

        :param event_label: (str) key for button to change
        :param color: (str) color to change to
        """
        self.event_buttons[event_label].change_background_color(color)

    def get_container_info(self, container_label):
        return self.containers[container_label].get_items()
    
    # Methods to be called by the process launching the GUI

    def configure_widgets(self):
        """ Configures all widgets in the queue

        Simply passes if the widget does not exist in the GUI
        """

        # Clear plot widgets
        for plot_widget in self._plots_to_remove:

            # Remove plot widget
            try:
                self._clear_plot(plot_widget)
                self._plots_to_remove.remove(plot_widget)
            except KeyError:
                pass

        # Assign plot widgets
        for plot_widget_params in self._plots_to_assign:

            # Unpack parameters
            plot_widget, plot_label, legend_widget = plot_widget_params

            # Assign plot to physical plot widget in GUI
            try:
                self._assign_plot(plot_widget, plot_label, legend_widget)
                self._plots_to_assign.remove(plot_widget_params)
            except KeyError:
                pass

        # Assign curves to plots
        for curve_params in self._curves_to_assign:

            # Unpack parameters
            plot_label, curve_label, error = curve_params

            # Assign curve to physical plot widget in GUI
            try:
                self._assign_curve(plot_label, curve_label, error)
                self._curves_to_assign.remove(curve_params)
            except KeyError:
                pass

        # Assign scalar widgets
        for scalar_params in self._scalars_to_assign:

            # Unpack parameters
            scalar_widget, scalar_label = scalar_params

            # Assign scalar to physical scalar widget in GUI
            try:
                self._assign_scalar(scalar_widget, scalar_label)
                self._scalars_to_assign.remove(scalar_params)
            except KeyError:
                pass

        # Assign label widgets
        for label_params in self._labels_to_assign:

            # Unpack parameters
            label_widget, label_label = label_params

            # Assign label to physical label widget in GUI
            try:
                self._assign_label(label_widget, label_label)
                self._labels_to_assign.remove(label_params)
            except KeyError:
                pass

        for curve_params in self._curves_to_remove:

            # Unpack parameters
            plot_label, curve_label = curve_params

            try:
                self._remove_curve(plot_label, curve_label)
                self._curves_to_remove.remove(curve_params)
            except KeyError:
                pass

        for event_button in self._buttons_to_assign:

            # Unpack parameters
            event_widget, event_label = event_button

            try:
                self._assign_event_button(event_widget, event_label)
                self._buttons_to_assign.remove(event_button)
            except KeyError:
                pass

        for cont in self._containers_to_assign:

            # Unpack parameters
            container_widget, container_label = cont

            try:
                self._assign_container(container_widget, container_label)
                self._containers_to_assign.remove(cont)
            except KeyError:
                pass
    
    def update_widgets(self):
        """ Updates all widgets on the physical GUI to current data"""

        # Update all plots
        for plot in self.plots.values():
            plot.update_output()

        # Update all scalars
        for scalar in self.scalars.values():
            scalar.update_output()

    def force_update(self):
        """ Forces the GUI to update.

        MUST be called in order for the GUI to be responsive, otherwise it freezes

        :return: 0 when complete
        """
        self._app.processEvents()
        return 0

    # Technical methods

    def _load_gui(self, gui_template=None, run=True):
        """ Loads a GUI template to the main window.

        Currently assumes all templates are in the directory given by the self._gui_directory attribute. If no
        gui_template is passed, the self._default_template is used. By default, this method also runs the GUI window.

        :param gui_template: name of the GUI template to use (str)
        :param run: whether or not to also run the GUI (bool)
        """

        if gui_template is None:
            gui_template = self._default_template

        # Check for proper formatting
        if not gui_template.endswith(".ui"):
            gui_template += ".ui"

        # Find path to GUI
        # Currently assumes all templates are in the directory given by the self._gui_directory attribute
        self._ui = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            self._gui_directory,
            gui_template
        )

        # Load UI
        try:
            uic.loadUi(self._ui, self)
        except FileNotFoundError:
            raise

        self._initialize_step_sizes()

        if run:
            self._run_gui()

    def _initialize_step_sizes(self):
        """ Initializes all QDoubleSpinBox step sizes based on step_sizes widget

        If step_sizes widget is not present, does nothing and initializes self.num_widgets to None
        """

        # Get all numerical widgets in the GUI
        self.num_widgets = [widget for widget in self.__dict__.values() if isinstance(widget, QtWidgets.QDoubleSpinBox)]
        try:

            # Exclude the step size setting
            self.num_widgets.remove(self.step_size)

            # Get current step size
            step_size = self.step_size.value()

            # Set all step sizes
            for widget in self.num_widgets:
                widget.setSingleStep(step_size)

            self.step_size.valueChanged.connect(self._update_step_sizes)

        # Handle case where step size was not in the GUI
        except AttributeError:
            self.num_widgets = None

    def _update_step_sizes(self):
        """ Updates all step sizes of QDoubleSpinBox widgets to current value of step_size widget """

        # Get current step size
        step_size = self.step_size.value()

        # Set all step sizes
        for widget in self.num_widgets:
            widget.setSingleStep(step_size)

    def _run_gui(self):
        """Runs the GUI. Displays the main window"""

        self.show()


    def _assign_plot(self, plot_widget, plot_label, legend_widget):
        """ Assigns a plot to a particular plot widget

        :param plot_widget: (str) name of the plot widget to use in the GUI window
        :param plot_label: (str) label to use for future referencing of the plot widget from the self.plots dictionary
        :param legend_widget: (str) name of the GraphicsView widget to use in the GUI window for this plot legend
        """

        # Assign plot
        self.plots[plot_label] = Plot(self, plot_widget, legend_widget)

        # Set title
        self.plots[plot_label].widget.setTitle(plot_label)

    def _clear_plot(self, plot_widget):
        """ Clears a plot and removes it from the self.plots attribute

        :param plot_widget: (str) name of the plot widget (instance of PYQTgraph) to clear
        """

        # Identify the plot widget
        widget_to_remove = getattr(self, plot_widget)

        # Find the relevant widget
        plot_to_delete = None
        for plot_label, plot in self.plots.items():
            if plot.widget is widget_to_remove:

                # Remove all curves
                curves_to_remove = []
                for curve_label in plot.curves:
                    curves_to_remove.append(curve_label)
                for curve_to_remove in curves_to_remove:
                    plot.remove_curve(curve_to_remove)

                # Remove plot from self.plots
                plot_to_delete = plot_label
        if plot_to_delete is not None:
            del self.plots[plot_to_delete]

    def _assign_curve(self, plot_label, curve_label, error=False):
        """ Assigns a curve to a plot

        :param plot_label: (str) label of the plot to assign
        :param curve_label: (str) label of curve to use for indexing in the self.plots[plot_label].curves dictionary
        :param error: (bool, optional) whether or not to use error bars
        """
        self.plots[plot_label].add_curve(curve_label, error)

    def _remove_curve(self, plot_label, curve_label):
        """ Removes a curve from a plot

        :param plot_label: (str) label of plot holding curve
        :param curve_label: (str) label of curve to remove
        """
        self.plots[plot_label].remove_curve(curve_label)

    def _assign_scalar(self, scalar_widget, scalar_label):
        """ Assigns scalar widget display in the GUI

        :param scalar_widget: (str) name of the scalar widget (e.g. QLCDNumber) in the GUI
        :param scalar_label: (str) label of the scalar widget
        """
        self.scalars[scalar_label] = Scalar(self, scalar_widget)

    def _assign_label(self, label_widget, label_label):
        """ Instantiates label object and assigns it to reference string

        :param label_widget: (str) name of the label widget (e.g. QLabel) in the GUI
        :param label_label: (str) reference string for the label
        """
        self.labels[label_label] = Label(self, label_widget)

    def _assign_event_button(self, event_widget, event_label):
        """ Assigns physical event button

        :param event_widget: (str) name of physical button widget on GUI
        :param event_label: (str) key name for reference to button
        """

        self.event_buttons[event_label] = EventButton(
            gui=self,
            event_widget=event_widget
        )

    def _assign_container(self, container_widget, container_label):
        """ Assigns physical ;ost

        :param container_widget: (str) name of physical container widget on GUI
        :param container_label: (str) key name for reference to the container
        """

        self.containers[container_label] = Container(
            gui=self,
            widget=container_widget
        )

class Plot:
    """ Class for plot widgets inside of a Window

    See also Curve class for specific curves, multiple of which can be assigned within a plot
    """

    # Semi-arbitrary list of colors to cycle through for plot data
    _color_list = ['#e6194b', '#3cb44b', '#ffe119', '#4363d8', '#f58231', '#911eb4', '#46f0f0', '#f032e6', '#bcf60c',
                   '#fabebe', '#008080', '#e6beff', '#9a6324', '#fffac8', '#800000', '#aaffc3', '#808000', '#ffd8b1',
                   '#000075', '#808080']

    def __init__(self, gui, plot_widget, legend_widget):
        """ Instantiates plot object for a plot inside of the GUI window

        See technical details about the legend implementation in the _set_legend() method

        :param gui: (Window) instance of the GUI window
        :param plot_widget: (str) name of the plot widget
        :param legend_widget: (str) name of the legend widget
        """

        # Assign the actual widget object, which is an instance of pyqtgraph.PlotWidget
        self.gui = gui
        self.widget = getattr(self.gui, plot_widget)

        self.curves = {}  # Holds all curves for this plot with associated curve label key
        self.legend = None  # Holds legendItem instance for plot legend

        # Set up legend
        self._set_legend(legend_widget=legend_widget)

    def add_curve(self, curve_label, error=False):
        """ Adds a curve to the plot

        TODO (if someone wants...): implement custom curve property override (e.g. custom color, line style, etc)

        :param curve_label: (str) curve label key to attach to this curve for future reference + naming purposes
        :param error: (bool, optional) whether or not to use error bars
        """

        # Add a new curve to self.curves dictionary for this plot
        self.curves[curve_label] = Curve(
            self.widget,
            pen=pg.mkPen(color=self._color_list[len(self.curves)]),  # Instantiate a Pen for curve properties
            error=error
        )

        # Configure legend
        self._add_to_legend(curve_label)

    def remove_curve(self, curve_label):
        """ Removes a curve from the plot

        :param curve_label: (str) name of curve to remove
        """

        # Remove from legend
        self._remove_from_legend(curve_label)

        # Remove from plot
        self.widget.removeItem(self.curves[curve_label].widget)

        # Remove from dictionary of curves
        del self.curves[curve_label]

    def update_output(self):
        """ Updates plot output to latest data"""

        for curve in self.curves.values():

            # Set data without error bars
            curve.widget.setData(
                curve.data
            )

            if curve.error_data is not None:
                curve.error.setData(
                    x=np.arange(len(curve.data)),
                    y=curve.data,
                    height=2*curve.error_data
                )

    # Technical methods

    def _set_legend(self, legend_widget):
        """ Sets the legend for a plot.

        Should be invoked once at the beginning when a plot is initialized. Requires a GraphicsView instance in order to
        serve as a container for the legend. This can be done by dragging a QGraphicsView object into the desired shape
        and location in the GUI, and promoting to a GraphicsView object with the pyqtgraph header. The legend is
        configured to sit in the top left corner of the box.

        :param legend_widget: (str) name of GraphicsView instance in the GUI window to store legend in
        """

        # Assign legend widget to plot_label key in self.legends
        self.legend = pg.LegendItem()

        # Assign legend widget to GraphicsView instance provided using a ViewBox object as an anchor
        # Simply tying the legend to the plot causes it to overlap with the data
        # Not sure this is the best way to solve this, but this solution enables .ui defined legend placement
        view_box = pg.ViewBox()
        getattr(self.gui, legend_widget).setCentralWidget(view_box)
        self.legend.setParentItem(view_box)
        self.legend.anchor((0, 0), (0, 0))

    def _add_to_legend(self, curve_label):
        """ Adds a curve to the legend

        :param curve_label: (str) key label of curve in self.curves dictionary to add to legend
        """

        # Add some space to the legend to prevent overlapping
        self.legend.addItem(self.curves[curve_label].widget, ' - '+curve_label)

    def _remove_from_legend(self, curve_label):
        """ Removes a curve from the legend

        :param curve_label: (str) label of curve to remove from legend
        """
        self.legend.removeItem(' - '+curve_label)


class Curve:
    """ Class for individual curves within a Plot """

    def __init__(self, plot_widget, pen=None, error=False):
        """ Instantiates a curve inside of a plot_widget

        :param plot_widget: (pg.PlotItem) instance of plot widget to add curve to
        :param pen: (pg.Pen) pen to use to draw curve
        :param error: (bool, optional) whether or not to add error bars to this plot
        """

        # Instantiate error bar plot
        self.widget = plot_widget.plot(pen=pen)

        if error:
            self.error = pg.ErrorBarItem(
                x=np.array([]),
                y=np.array([]),
                pen=pen
            )
            plot_widget.addItem(self.error)
            self.error_data = np.array([])

        # Instantiate new curve in the plot without error bars
        else:
            self.widget = plot_widget.plot(pen=pen)
            self.error = None
            self.error_data = None

        self.data = np.array([])

    def set_curve_data(self, data, error=None):
        """ Stores data to a new curve

        :param data: (np.array) either a 1D (only y-axis) or 2D (x-axis, y-axis) numpy array
        :param error: (np.array, optional) 1D array for error bars (standard deviation)
        """

        self.data = data

        if self.error_data is not None:
            self.error_data = error


class Scalar:
    """ A scalar display object (e.g. a number or boolean)

    Currently supports any instance of QAbstractButton (booleans such as pushbuttons and checkboxes), QAbstractSpinbox
    (numerical displays/inputs) and QLCDNumber (swanky digital number displays)

    Also supports two-way functionality (not just data display). See module doctstring for details
    """

    def __init__(self, gui, scalar_widget):
        """ Initializes the scalar object

        By default, activates the widget so it can be updated via an external script and is effectively blocked from
        user input. Call the deactivate() method in order to deactivate and enable GUI input.

        :param gui: (Window) instance of the GUI window class containing the number widget
        :param scalar_widget: (str) name of the widget for reference
        """

        self.data = None
        self._use_data = True  # Whether or not the widget is active

        # Get actual widget instance
        self.widget = getattr(gui, scalar_widget)

    def set_data(self, data):
        """ Sets the value of the scalar internally

        :param data: number to set to
        """
        self.data = data

    def update_output(self):
        """ Updates the physical output of the scalar widget

        Checks if the widget is active, in which case it sets the output to self.data. If not, it just maintains the
        current output of the GUI.
        """

        # Check if active
        if self._use_data:

            # Set the state, checking first whether it is a boolean display or not
            if isinstance(self.data, bool):
                # Check if the script has updated and implement
                self.widget.setChecked(self.data)

            # Handle numerical widgets
            elif self.data is not None:
                # Now if it is simply a QLCDButton, it should have the display method
                try:
                    if self.widget.value() != self.data:
                        display_str = '%.6f' % self.data
                        self.widget.display(display_str)

                # In case client disconnects
                except EOFError:
                    pass

                # If it is an input numerical display, update!
                except AttributeError:
                    try:
                        self.widget.setValue(self.data)
                    # In case client disconnects

                    except EOFError:
                        pass

                    # In case it is some other funky widget we haven't accounted for
                    # Future addition to scalar implementation can go inside here
                    except AttributeError:
                        pass

    def get_data(self):
        """ Gets the current data from the GUI widget

        :return: current data (either a bool or a double)
        """

        # First try the boolean implementation
        try:
            data = self.widget.isChecked()

        # Next try the numerical (works for QLCDNumber and QAbstractSpinBox
        except AttributeError:
            data = self.widget.value()
        return data

    def activate(self):
        """ Tells a scalar to use its data, locking GUI editing and updating from the script"""
        self._use_data = True

    def deactivate(self):
        """ Tells a scalar not to use its data, enables free GUI editing"""
        self._use_data = False


class Label:
    """ A text label display object. Currently supports QLabel"""

    def __init__(self, gui, label_widget):
        """ Constructor for label object

        :param gui: (Window) instance of GUI window class containing the label object
        :param label_widget: (str) name of the widget for reference
        """

        self.text = None
        self.widget = getattr(gui, label_widget)

    def set_label(self, label_text=''):
        """ Sets label text

        :param label_text: (str, optional) text string to set the label to
        """

        self.text = label_text
        self.widget.setText(label_text)

    def get_label(self):
        """ Returns label text """
        
        try:
            
            # Ordinary labels 
            return self.widget.text()
        except AttributeError:

            # Fancy labels
            return self.widget.toPlainText()


class EventButton:
    """ Class for event pushbuttons to be tied to some script functions externally """

    def __init__(self, gui, event_widget):
        """ Instantiates event button """

        self.was_pushed = False  # Keeps track of whether the button has been pushed
        self.widget = getattr(gui, event_widget)  # Get physical widget instance

        self.disabled = False  # Check if button is disabled

        # Connect event to flag raising
        self.widget.pressed.connect(self.button_pressed)

    def button_pressed(self):
        """ Raises flag when button is pushed """

        self.was_pushed = True

    def reset_button(self):
        """ Resets pushed state to False """

        self.was_pushed = False

    def change_background_color(self, color):
        """ Changes background_color of button"""

        self.widget.setStyleSheet(f"background-color: {color}")

    def get_state(self):
        """ Returns whether or not the button has been pressed and resets button state

        :return: bool self._was_pushed
        """

        result = copy.deepcopy(self.was_pushed)
        self.reset_button()
        return result


class Container:
    """ Class for generic containers with elements added within

    Idea being that all that needs to be referenced is the top level
    and methods here can be invoked to get information about containing elements

    Only QListWidget supported
    """

    def __init__(self, gui, widget):
        """ Instantiates event button """

        self.widget = getattr(gui, widget)  # Get physical widget instance

    def get_items(self):
        """ Returns all QListWidget items and tooltips as a dictionary """
        
        item_info = {}
        for index in range(self.widget.count()):

            # Get the current item and store its name and tooltip text
            current_item = self.widget.item(index)
            item_info[current_item.text()] = current_item.toolTip()

        return item_info



import abc

class MultiTraceInterface(abc.ABC):

    @abc.abstractmethod
    def set_data(self, x_ar=None, y_ar=None, ind=0, noise=None):
        pass

    @abc.abstractmethod
    def set_lbls(self, x_str=None, y_str=None):
        pass

    @abc.abstractmethod
    def show(self):
        pass


class TraceInterface(abc.ABC):

    @abc.abstractmethod
    def set_data(self, x_ar=None, y_ar=None):
        pass

    @abc.abstractmethod
    def append_data(self, x_ar=None, y_ar=None):
        pass

    @abc.abstractmethod
    def set_lbls(self, x_str=None, y_str=None):
        pass


class HeatMapInterface(abc.ABC):

    @abc.abstractmethod
    def set_data(self, x_ar=None, y_ar=None, z_ar=None):
        pass

    @abc.abstractmethod
    def append_row(self, y_val=None, z_ar=None):
        pass

    @abc.abstractmethod
    def append_col(self, x_val=None, z_ar=None):
        pass

    @abc.abstractmethod
    def set_lbls(self, x_str=None, y_str=None):
        pass


class PBarInterface(abc.ABC):

    @abc.abstractmethod
    def set_value(self, value):
        pass




''' Configures a Staticline instance to use a NIDaqmx output'''

import socket

import pylabnet.hardware.ni_daqs.nidaqmx_card as nidaqmx
import pylabnet.hardware.staticline.staticline as staticline

from pylabnet.network.client_server.staticline import Service, Client

from pylabnet.network.core.generic_server import GenericServer

# Parameters
NI_DEVICE_NAME = 'cDAQ1Mod1'


def launch(**kwargs):
    """ Connects to a NI DAQ as staticline

    :param kwargs: (dict) containing relevant kwargs
        :logger: instance of LogClient for logging purposes
        :port: (int) port number for the Cnt Monitor server
    """

    staticline_logger = kwargs['logger']
    daq = nidaqmx.Driver(device_name=NI_DEVICE_NAME, logger=staticline_logger)

    test_staticline = staticline.Driver(
        name='Test staticline',
        logger=kwargs['logger'],
        hardware_module=daq,
        ao_output='ao2',
        down_voltage=0,
        up_voltage=3.3,
    )

    # Instantiate Server
    # Staticline server
    staticline_service = Service()
    staticline_service.assign_module(module=test_staticline)
    staticline_service.assign_logger(logger=staticline_logger)
    staticline_service_server = GenericServer(
        service=staticline_service,
        host=socket.gethostbyname(socket.gethostname()),
        port=kwargs['port']
    )

    staticline_service_server.start()



import numpy as np


# Base classes ----------------------------------------------------------------

class PulseBase:
    """ Base class for Pulse objects
    """

    def __init__(self, ch, dur, t0=0):
        """ Base class for Pulse object

        :param ch: (str) channel name
        :param dur: (numeric) duration of the pulse
        :param t0: (opt, numeric) position of the pulse beginning
        """

        self.ch = ch
        self.t0 = t0
        self.dur = dur


class DfltPulseBase:
    """ Base class for DfltPulse objects
    """

    def __eq__(self, other):
        raise NotImplementedError(
            'Every default pulse class has to define __eq__() method. \n'
            'Reason: during insert_pb() call, a check for conflicts between '
            'default pulse objects on the overlapping channels is performed'
            'by direct comparison operation.'
        )


# Pulse classes ---------------------------------------------------------------

class PTrue(PulseBase):
    """ Pulse: Boolean True
    """

    def __init__(self, ch, dur, t0=0):
        super().__init__(ch=ch, dur=dur, t0=t0)

    def __str__(self):
        return 'True'

    def get_value(self, t_ar):
        """ Returns array of samples

        :param t_ar: (numpy.array) array of time points
        :return: (numpy.array(dtype=bool)) array of samples
        """

        return np.full(len(t_ar), True)


class PFalse(PulseBase):
    """ Pulse: Boolean False
    """

    def __init__(self, ch, dur, t0=0):
        super().__init__(ch=ch, dur=dur, t0=t0)

    def __str__(self):
        return 'False'

    def get_value(self, t_ar):
        """ Returns array of samples

        :param t_ar: (numpy.array) array of time points
        :return: (numpy.array(dtype=bool)) array of samples
        """

        return np.full(len(t_ar), False)


class PSin(PulseBase):
    """ Pulse: Sine
    """

    def __init__(self, ch, dur, t0=0, amp=0, freq=0, ph=0):
        """ Construct Sin Pulse object

        :param ch: (str) channel name
        :param dur: (numeric) duration of the pulse
        :param t0: (opt, numeric) position of the pulse beginning
        :param amp: (opt, np.float32) amplitude (zero-to-peak)
        :param freq: (opt, np.float32) frequency (linear, without 2*pi)
        :param ph: (opt, np.float32) phase (in degrees)
        """

        super().__init__(ch=ch, dur=dur, t0=t0)

        self._amp = amp
        self._freq = freq
        self._ph = ph

    def __str__(self):
        ret_str = 'Sin(amp={:.2e} freq={:.2e} ph={:.2f})' \
                  ''.format(self._amp, self._freq, self._ph)
        return ret_str

    def get_value(self, t_ar):
        """ Returns array of samples

        :param t_ar: (numpy.array) array of time points
        :return: (numpy.array(dtype=np.float32)) array of samples
        """

        t_ar = np.array(t_ar, dtype=np.float32)
        ret_ar = self._amp * np.sin(
            2*np.pi*self._freq*t_ar + np.pi*self._ph/180
        )
        return ret_ar


class PConst(PulseBase):
    """ Pulse: Constant
    """

    def __init__(self, ch, dur, t0=0, val=0.0):
        """ Construct Constant Pulse

        :param ch: (str) channel name
        :param dur: (numeric) duration of the pulse
        :param t0: (opt, numeric) position of the pulse beginning
        :param val: (opt, np.float32) constant value
        """

        super().__init__(ch=ch, dur=dur, t0=t0)
        self._val = val

    def __str__(self):
        return 'Const(val={})'.format(self._val)

    def get_value(self, t_ar):
        """ Returns array of samples

        :param t_ar: (numpy.array) array of time points
        :return: (numpy.array(dtype=np.float32)) array of samples
        """

        t_ar_len = len(t_ar)
        ret_ar = np.full(t_ar_len, self._val, dtype=np.float32)

        return ret_ar


# Default Pulse classes -------------------------------------------------------

class DFalse(DfltPulseBase):
    """ DfltPulse: Boolean False
    """

    def __eq__(self, other):
        if not isinstance(other, DFalse):
            return NotImplemented

        return True

    def __str__(self):
        return 'False'

    @staticmethod
    def get_value(t_ar):
        """ Returns array of samples

        :param t_ar: (numpy.array) array of time points
        :return: (numpy.array(dtype=bool)) array of samples
        """

        t_ar_len = len(t_ar)
        return np.full(t_ar_len, False)


class DTrue(DfltPulseBase):
    """ DfltPulse: Boolea True
    """

    def __eq__(self, other):
        if not isinstance(other, DTrue):
            return NotImplemented

        return True

    def __str__(self):
        return 'True'

    @staticmethod
    def get_value(t_ar):
        """ Returns array of samples

        :param t_ar: (numpy.array) array of time points
        :return: (numpy.array(dtype=bool)) array of samples
        """

        t_ar_len = len(t_ar)
        return np.full(t_ar_len, True)


class DConst(DfltPulseBase):
    """ DfltPulse: Float Constant
    """

    def __init__(self, val=0.0):
        """ Construct Constant DfltPulse object

        :param val: (numpy.float32) value of the constant
        """

        self._val = val

    def __eq__(self, other):
        if not isinstance(other, DConst):
            return NotImplemented

        if self._val == other._val:
            return True
        else:
            return False

    def __str__(self):
        return 'Const(val={})'.format(self._val)

    def get_value(self, t_ar):
        """ Returns array of samples

        :param t_ar: (numpy.array) array of time points
        :return: (numpy.array(dtype=np.float32)) array of samples
        """

        t_ar_len = len(t_ar)
        ret_ar = np.full(t_ar_len, self._val, dtype=np.float32)

        return ret_ar






import abc


class MWSrcInterface(abc.ABC):

    @abc.abstractmethod
    def activate_interface(self):
        pass

    # Output control

    @abc.abstractmethod
    def on(self):
        pass

    @abc.abstractmethod
    def off(self):
        pass

    @abc.abstractmethod
    def get_status(self):
        pass

    # Power

    @abc.abstractmethod
    def get_pwr(self):
        pass

    @abc.abstractmethod
    def set_pwr(self, pwr):
        pass

    # Frequency

    @abc.abstractmethod
    def get_freq(self):
        pass

    @abc.abstractmethod
    def set_freq(self, freq):
        pass

    @abc.abstractmethod
    def set_freq_swp(self, start, stop, n_pts):
        pass

    @abc.abstractmethod
    def reset_swp_pos(self):
        pass

    @abc.abstractmethod
    def get_mode(self):
        pass


class MWSrcError(Exception):
    pass




# fw102c.py
#
# Thorlabs FW102C - Six-Position Motorized Filter Wheel - Python interface

# Gilles Simond <gilles.simond@unige.ch>	

# who       when        what
# --------  ----------  -------------------------------------------------
# gsimond   20140922    modified from Adrien.Deline version
#

import io,re,sys
from serial import Serial, SerialException

class FW102C(object):
    """
       Class to control the ThorLabs FW102C filter wheel
       
          fwl = Thorlabs.FW102C(port='COM5')
          fwl.help()
          fwl.command('pos=5')
          fwl.query('pos?')
          fwl.close()
          
       The following table describes all of the available commands and queries:
        *idn?     Get ID: Returns the model number and firmware version
        pos=n     Moves the wheel to filter position n
        pos?      Get current Position
        pcount=n  Set Position Count: Sets the wheel type where n is 6 or 12
        pcount?   Get Position Count: Returns the wheel type
        trig=0    Sets the external trigger to the input mode
        trig=1    Sets the external trigger to the output mode
        trig?     Get current Trigger Mode
        speed=0   Sets the move profile to slow speed
        speed=1   Sets the move profile to high speed
        speed?    Returns the move profile mode
        sensors=0 Sensors turn off when wheel is idle to eliminate stray light
        sensors=1 Sensors remain active
        sensors?  Get current Sensor Mode
        baud=0    Sets the baud rate to 9600
        baud=1    Sets the baud rate to 115200
        baud?     Returns the baud rate where 0 = 9600 and 1 = 115200
        save      This will save all the settings as default on power up
        
    """
    isOpen   = False
    devInfo  = None

    def __init__(self, port, logger):

        self.log = logger

        try:
            self._fw = Serial(port=port, baudrate=115200,
                                  bytesize=8, parity='N', stopbits=1,
                                  timeout=1, xonxoff=0, rtscts=0)
        except  SerialException as ex:
            self.log.error('Port {0} is unavailable: {1}'.format(port, ex))
            return
        except  OSError as ex:
            self.log.error('Port {0} is unavailable: {1}'.format(port, ex))
            return
        self._sio = io.TextIOWrapper(io.BufferedRWPair(self._fw, self._fw, 1),
                       newline=None, encoding='ascii')


        
        self._sio.write(str('*idn?\r'))
        self.devInfo = self._sio.readlines(2048)[1][:-1]

        
        self.log.info(self.devInfo)
        self._sio.write(str('pos?\r'))
        self.pos = self._sio.readlines(2048)[1][:-1]

        self.log.info ('position={}'.format(self.pos))
        self._sio.write(str('pcount?\r'))
        self.pcount = self._sio.readlines(2048)[1][:-1]

        self.log.info ('pcount={}'.format(self.pcount))
        self._sio.write(str('trig?\r'))
        self.trig = self._sio.readlines(2048)[1][:-1]

        self.log.info ('trig={}'.format(self.trig))
        self._sio.write(str('speed?\r'))
        self.speed = self._sio.readlines(2048)[1][:-1]

        self.log.info ('speed= {}'.format(self.speed))
        self._sio.write(str('sensors?\r'))
        self.sensors = self._sio.readlines(2048)[1][:-1]

        self.log.info ('sensors={}'.format(self.sensors))
        self._sio.write(str('baud?\r'))
        self.baud = self._sio.readlines(2048)[1][:-1]

        if self.baud: 
            self.baud = 115200
        else: 
            self.baud = 9600
        self.log.info ('baud= {}'.format(self.baud))
				
        self._sio.flush()
        self.isOpen  = True
    # end def __init__
    
    def help(self):
        print(self.__doc__)
    # end def help
    
    def close(self):
        if not self.isOpen:
            print("Close error: Device not open")
            return "ERROR"
        #end if
        
        self._fw.close()
        self.isOpen = False
        return "OK"
    # end def close
	
    def query(self, cmdstr):
        """
           Send query, get and return answer
        """
        if not self.isOpen:
            self.log.error("Query error: Device not open")
            return "DEVICE NOT OPEN"
        #end if
        
        ans = 'ERROR'
        self._sio.flush()
        res = self._sio.write(str(cmdstr+'\r'))
        if res:
            ans = self._sio.readlines(2048)[1][:-1]
        #print 'queryans=',repr(ans)
        return ans
    # end def query
	
    def command(self, cmdstr):
        """
           Send command, check for error, send query to check and return answer
           If no error, answer value should be equal to command argument value
        """
        if not self.isOpen:
            self.log.error("Command error: Device not open")
            return "DEVICE NOT OPEN"
        #end if
        
        ans = 'ERROR'
        self._sio.flush()
        cmd = cmdstr.split('=')[0]
        res = self._sio.write(str(cmdstr+'\r'))
        ans = self._sio.readlines(2048)
        regerr = re.compile("Command error.*")
        errors = [m.group(0) for l in ans for m in [regerr.search(l)] if m]
        #print 'res=',repr(res),'ans=',repr(ans),cmd
        if len(errors) > 0:
            return errors[0]
        ans = self.query(cmd+'?')
        #print 'ans=',repr(ans),cmd+'?'
        return ans
    # end def command
	
    def getinfo(self):
        if not self.isOpen:
            self.log.error("Getinfo error: Device not open")
            return "DEVICE NOT OPEN"
        #end if
        
        return self.devInfo
    # end def getinfo
	
#end class FW102C

# # Class test, when called directly
# if __name__ == "__main__":
#     fwl = FW102C(port='COM10')
#     if not fwl.isOpen:
#         print("FWL INIT FAILED")
#         sys.exit(2)
#     print('**info',fwl.getinfo())
#     print( '**idn?',fwl.query('*idn?'))
    
#     print( '**pos=5',fwl.command('pos=5'))
#     print( '**pos?',fwl.query('pos?'))
#     print( '**pos=7',fwl.command('pos=7'))
#     print( '**pos?',fwl.query('pos?'))
#     print( '**pos=3',fwl.command('pos=3'))
#     print( '**pos?',fwl.query('pos?'))
#     print( '**pos=6',fwl.command('pos=6'))
#     print( '**pos?',fwl.query('pos?'))
#     print( '**qwzs=3',fwl.command('qwz=3'))
#     print( '**pos?',fwl.query('pos?'))
#     print( fwl.close())


from pylabnet.gui.pyqt.gui_handler import GUIHandler
from pylabnet.utils.helper_methods import unpack_launcher

import time


class StaticLineGUIGeneric():
    """Generic Static Line GUI

    :gui_client: (object)
        GUI client to be called.
    :logger_client: (object)
        Logger client used for error logging in  @handle_gui_errors
        decorator.
    :staticline: (object)
        Instance of staticline.Driver
    """

    def __init__(
        self,
        gui_client,
        staticline_client,
        logger_client
    ):

        # Instanciate gui handler
        self.gui_handler = GUIHandler(gui_client, logger_client)

        # store shutter client
        self.staticline = staticline_client
        # self.staticline_HDAWG = staticline_hdawg

    def initialize_button(
        self,
        button_widget='button_label',
        button_widget_label='button_label_label',
        button_up_widget='up_button',
        button_up_widget_label='up_button_label',
        button_down_widget='down_button',
        button_down_widget_label='down_button_label',

    ):
        """Initialize label to staticline name, assignes labels."""

        label_widgets = [
            button_widget,
        ]

        label_labels = [
            button_widget_label,

        ]

        button_widgets = [
            button_up_widget,
            button_down_widget,
        ]

        button_labels = [
            button_up_widget_label,
            button_down_widget_label,
        ]

        # Assign label labels
        for widget, label in zip(label_widgets, label_labels):
            self.gui_handler.assign_label(widget, label)

        # Assign buttons
        for widget, label in zip(button_widgets, button_labels):
            self.gui_handler.assign_event_button(widget, label)

        # Retrieve staticline names
        staticline_name = self.staticline.get_name()

        # Set value of label staticline names.
        self.gui_handler.set_label(staticline_name, button_widget_label)

        # Store label name
        self.button_up_widget_label = button_up_widget_label
        self.button_down_widget_label = button_down_widget_label

    def check_buttons(self):
        """ Checks state of buttons and sets shutter accordingly"""
        if self.gui_handler.was_button_pressed(event_label=self.button_up_widget_label):
            self.staticline.up()
        elif self.gui_handler.was_button_pressed(event_label=self.button_down_widget_label):
            self.staticline.down()

    def run(self):

        # Initialize button
        self.initialize_button()

        # Mark running flag
        self.gui_handler.is_running = True
        while self.gui_handler.is_running:
            self.check_buttons()
            time.sleep(0.02)


def launch(**kwargs):
    """Launches the script."""

    logger, loghost, logport, clients, guis, params = unpack_launcher(**kwargs)

    # Instantiate StaticLineGUIGeneric
    try:
        staticline_gui = StaticLineGUIGeneric(
            staticline_client=clients['staticline_nidaqmx'], gui_client=guis['staticline_generic'], logger_client=logger
        )
    except KeyError:
        logger.error('Please make sure the module names for required servers and GUIS are correct.')

    # Run
    staticline_gui.run()



import rpyc


class ClientBase:
    def __init__(self, host, port):

        # Internal vars to store server info
        self._host = ''
        self._port = 0

        # Internal vars to store refs to server
        self._connection = None
        self._service = None

        # Connect to server
        self.connect(host=host, port=port)

    def connect(self, host='place_holder', port=-1):

        # Update server address if new values are given
        if host != 'place_holder':
            self._host = host
        if port != -1:
            self._port = port

        # Clean-up old connection if it exists
        if self._connection is not None or self._service is not None:
            try:
                self._connection.close()
            except:
                pass

            self._connection = None
            self._service = None

        # Connect to server
        try:
            self._connection = rpyc.connect(
                host=self._host,
                port=self._port,
                config={
                    'allow_public_attrs': True,
                    'sync_request_timeout': 300
                }
            )
            self._service = self._connection.root

            return 0

        except Exception as exc_obj:
            self._connection = None
            self._service = None
            raise exc_obj



import plotly.graph_objs as go
from pylabnet.gui.output_interface import MultiTraceInterface, TraceInterface, HeatMapInterface, PBarInterface
import ipywidgets as iwdgt
from IPython.display import display
import copy
import numpy as np


class SingleTraceFig(TraceInterface):

    def __init__(self, mode='lines', title_str=None):

        self._fig = go.FigureWidget(data=[])

        self._fig.add_scatter(
            x=[],
            y=[],
            mode=mode
        )

        if title_str is not None:
            self._fig.layout.update(title=title_str)

        self._x_ar = np.array([])
        self._y_ar = np.array([])

    def set_data(self, x_ar=None, y_ar=None):

        # Set x_ar
        if x_ar is not None:
            self._x_ar = np.array(
                copy.deepcopy(x_ar)
            )

        # Set y_ar
        if y_ar is not None:
            self._y_ar = np.array(
                copy.deepcopy(y_ar)
            )

        # Update figure
        with self._fig.batch_update():
            self._fig.data[0].x = self._x_ar
            self._fig.data[0].y = self._y_ar

    def append_data(self, x_ar=None, y_ar=None):

        # Append new data to internal arrays
        if x_ar is not None:
            self._x_ar = np.append(self._x_ar, x_ar)

        if y_ar is not None:
            self._y_ar = np.append(self._y_ar, y_ar)

        # Apply changes to the figure
        if x_ar is not None or y_ar is not None:
            with self._fig.batch_update():
                self._fig.data[0].x = self._x_ar
                self._fig.data[0].y = self._y_ar

    def set_lbls(self, x_str=None, y_str=None):

        if x_str is not None:
            self._fig.layout.xaxis.title = x_str

        if y_str is not None:
            self._fig.layout.yaxis.title = y_str

    def show(self):
        display(self._fig)


class MultiTraceFig(MultiTraceInterface):

    def __init__(self, title_str=None, ch_names=None, shot_noise=False, legend_orientation=None):
        """ MultiTraceFig constructor

        :param title_str: string to display for plot title
        :param ch_names: list of channel name strings for multi
            -channel plot
        :param shot_noise: boolean of whether or not to plot
            shot noise on individual traces
        """

        self._fig = go.FigureWidget(data=[])
        self._shot_noise = shot_noise
        self._ch_names = ch_names

        if self._ch_names is not None:

            # Initialize all channels if given
            self._num_ch = 0
            self._allocate_arrays()

        else:
            self._num_ch = 0

        if title_str is not None:
            self._fig.layout.update(title=title_str)

        if legend_orientation is not None:
            self._fig.update_layout(
                legend=dict(x=0, y=1.1),
                legend_orientation=legend_orientation
            )

    def set_data(self, x_ar=None, y_ar=None, ind=0, noise=None):
        """ Sets data to MultiTraceFig instance

        :param x_ar: np array of x-axis data values
        :param y_ar: np array of y-axis data values
        :param ind: index of channels to assign data to
            (index starts from 0)
        :param noise: np array of y-axis error bar sizes
        """

        # First check that data array at particular index has
        # already been allocated, and if not, allocate the
        # previous data arrays
        to_allocate = ind - self._num_ch
        if to_allocate >= 0:
            self._allocate_arrays(
                num_arrays=to_allocate+1
            )

        # Update figure
        if x_ar is not None and y_ar is not None:

            if self._shot_noise:

                if noise is None:
                    noise = 0*y_ar

                # Input shot noise
                x_rev = x_ar[::-1]
                y_upper = y_ar + noise
                y_lower = y_ar - noise
                y_lower = y_lower[::-1]
                with self._fig.batch_update():
                    self._fig.data[ind*2].x = np.hstack((x_ar, x_rev))
                    self._fig.data[ind*2].y = np.hstack((y_upper, y_lower))

                # Now data
                with self._fig.batch_update():
                    self._fig.data[2*ind+1].x = x_ar
                    self._fig.data[2*ind+1].y = y_ar

            else:
                with self._fig.batch_update():
                    self._fig.data[ind].x = x_ar
                    self._fig.data[ind].y = y_ar

    def set_lbls(self, x_str=None, y_str=None):
        """

        :param x_str: x_axis label
        :param y_str: y_axis label
        """

        # Set titles of x and y axes as desired
        if x_str is not None:
            self._fig.layout.xaxis.title = x_str

        if y_str is not None:
            self._fig.layout.yaxis.title = y_str

    def show(self):
        display(self._fig)
        # self._fig.show(renderer="iframe")

    # Technical methods

    def _allocate_arrays(self, num_arrays=None):

        # Color-list
        dflt_plotly_colors = [
            'rgb(31, 119, 180)', 'rgb(255, 127, 14)',
            'rgb(44, 160, 44)', 'rgb(214, 39, 40)',
            'rgb(148, 103, 189)', 'rgb(140, 86, 75)',
            'rgb(227, 119, 194)', 'rgb(127, 127, 127)',
            'rgb(188, 189, 34)', 'rgb(23, 190, 207)'
        ]
        dflt_plotly_colors_lo = [
            'rgba(31, 119, 180, 0.2)', 'rgba(255, 127, 14, 0.2)',
            'rgba(44, 160, 44, 0.2)', 'rgba(214, 39, 40, 0.2)',
            'rgba(148, 103, 189, 0.2)', 'rgba(140, 86, 75, 0.2)',
            'rgba(227, 119, 194, 0.2)', 'rgba(127, 127, 127, 0.2)',
            'rgba(188, 189, 34, 0.2)', 'rgba(23, 190, 207, 0.2)'
        ]

        # If a certain number of arrays to be allocated is not assigned,
        # just allocate the number as given by number of elements in 
        if num_arrays is not None:
            for index in range(num_arrays):
                if self._shot_noise:

                    # Generate additional scatter for shot noise bounds
                    self._fig.add_scatter(
                        x=[],
                        y=[],
                        name='ch'+str(index),
                        showlegend=False,
                        fill='tozerox',
                        line=dict(color='rgba(255,255,255,0)'),
                        fillcolor=dflt_plotly_colors_lo[index]
                    )
                    self._fig.add_scatter(
                        x=[],
                        y=[],
                        mode='lines',
                        name='ch'+str(index),
                        line=dict(color=dflt_plotly_colors[index])
                    )
                else:
                    self._fig.add_scatter(
                        x=[],
                        y=[],
                        mode='lines',
                        name='ch'+str(index)
                    )
                self._num_ch += 1

        # Otherwise just use the self._ch_list to allocate arrays
        else:
            for index, channel in enumerate(self._ch_names):
                if self._shot_noise:

                    # Generate additional scatter for shot noise bounds
                    self._fig.add_scatter(
                        x=[],
                        y=[],
                        name=channel,
                        showlegend=False,
                        fill='tozerox',
                        line=dict(color='rgba(255,255,255,0)'),
                        fillcolor=dflt_plotly_colors_lo[index]
                    )
                    self._fig.add_scatter(
                        x=[],
                        y=[],
                        mode='lines',
                        name=channel,
                        line=dict(color=dflt_plotly_colors[index])
                    )
                else:
                    self._fig.add_scatter(
                        x=[],
                        y=[],
                        mode='lines',
                        name=channel
                    )
                self._num_ch += 1


class HeatMapFig(HeatMapInterface):

    def __init__(self, title_str=None):

        self._fig = go.FigureWidget(data=[])

        self._fig.add_heatmap(
            x=[],
            y=[],
            z=[[], []]
        )

        if title_str is not None:
            self._fig.layout.title = title_str

        self._x_ar = np.array([], dtype=float)
        self._y_ar = np.array([], dtype=float)
        self._z_ar = np.array([], dtype=float)

    def set_data(self, x_ar=None, y_ar=None, z_ar=None):

        # Set x_ar
        if x_ar is not None:
            self._x_ar = np.array(
                copy.deepcopy(x_ar)
            )

        # Set y_ar
        if y_ar is not None:
            self._y_ar = np.array(
                copy.deepcopy(y_ar)
            )

        # Set z_ar
        if z_ar is not None:
            self._z_ar = np.array(
                copy.deepcopy(z_ar)
            )

        # Update figure
        if x_ar is not None or y_ar is not None or z_ar is not None:
            with self._fig.batch_update():
                self._fig.data[0].x = self._x_ar
                self._fig.data[0].y = self._y_ar
                self._fig.data[0].z = self._z_ar

    def append_row(self, y_val=None, z_ar=None):

        if y_val is not None:
            self._y_ar = np.append(self._y_ar, y_val)

        if z_ar is not None:
            self._z_ar = np.append(
                self._z_ar,
                [z_ar],  # wrap into [] to make 2D array
                axis=0
            )

        # Apply changes to the figure
        if y_val is not None or z_ar is not None:
            with self._fig.batch_update():
                self._fig.data[0].y = self._y_ar
                self._fig.data[0].z = self._z_ar

    def append_col(self, x_val=None, z_ar=None):

        if x_val is not None:
            self._x_ar = np.append(self._x_ar, x_val)

        if z_ar is not None:
            self._z_ar = np.append(
                self._z_ar,
                np.transpose([z_ar]),
                axis=1
            )

        # Apply changes to the figure
        if x_val is not None or z_ar is not None:
            with self._fig.batch_update():
                self._fig.data[0].x = self._x_ar
                self._fig.data[0].z = self._z_ar

    def set_lbls(self, x_str=None, y_str=None):

        if x_str is not None:
            self._fig.layout.xaxis.title = x_str

        if y_str is not None:
            self._fig.layout.yaxis.title = y_str

    def show(self):
        display(self._fig)


class PBar(PBarInterface):

    def __init__(self, value=0):
        self._p_bar = iwdgt.FloatProgress()

        self._p_bar.min = 0
        self._p_bar.max = 100

        self.set_value(value=value)

    def set_value(self, value):
        self._p_bar.value = value
        self._p_bar.description = '{:.0f} %'.format(value)

    def show(self):
        display(self._p_bar)


class ComboTraceHMapPBar:

    def __init__(self, trace_title=None, hm_title=None):

        self.trace_fig = SingleTraceFig(title_str=trace_title)
        self.hm_fig = HeatMapFig(title_str=hm_title)
        self.p_bar = PBar()

        self._grid = iwdgt.VBox(
            [
                iwdgt.HBox([self.trace_fig._fig, self.hm_fig._fig]),
                self.p_bar._p_bar
            ]
        )

    def show(self):
        display(self._grid)






from pylabnet.hardware.interface.mw_src import MWSrcInterface
from pylabnet.hardware.interface.gated_ctr import GatedCtrInterface
from pylabnet.hardware.interface.simple_p_gen import SimplePGenInterface
from pylabnet.gui.output_interface import TraceInterface, HeatMapInterface, PBarInterface
import pylabnet.utils.pulseblock.pulse as po
import pylabnet.utils.pulseblock.pulse_block as pb
import numpy as np
import time

# TODO: add readout window and refocus as parameters


class Rabi:
    def __init__(self):

        # Hardware instances
        self._mw_src = None
        self._gated_ctr = None
        self._p_gen = None
        self._rfcs = None

        # GUI instances
        self._1d_trace = None
        self._2d_heat_map = None
        self._p_bar = None

        # Parameters
        #  - Rabi measurement
        self._freq = None
        self._pwr = None

        self._tau_start = None
        self._tau_stop = None
        self._n_pts = None

        #  - repetitions
        self._n_hrdw_reps = None
        self._n_soft_reps = None

        #  - PulseBlock params
        self._aom_delay = None

        # Internal data arrays
        self.tau_ar = None  # 1D numpy float array
        self.avg_ar = None  # 1D numpy float array
        self.raw_ar = None  # 2D numpy float array

        # Current software-rep index
        self._is_running = False
        self._soft_rep_idx = 0

        # Approximate duration of full RabiPB
        # [for gated_ctr.get_count_ar() timeout]
        self._rabi_pb_dur = -1

    # User interface methods

    def set_hardware(self, mw_src, gated_ctr, p_gen, rfcs=None):

        # Sanity checks
        if not isinstance(mw_src, MWSrcInterface):
            raise TypeError(
                'set_hardware(): given mw_src instance does not inherit from MWSrcInterface'
            )
        if not isinstance(gated_ctr, GatedCtrInterface):
            raise TypeError(
                'set_hardware(): given gated_ctr instance does not inherit from GatedCtrInterface'
            )
        if not isinstance(p_gen, SimplePGenInterface):
            raise TypeError(
                'set_hardware(): given p_gen instance does not inherit from SimplePGenInterface'
            )

        # Save references in internal variables
        self._mw_src = mw_src
        self._gated_ctr = gated_ctr
        self._p_gen = p_gen

        if rfcs is not None:
            self._rfcs = rfcs

        return 0

    def set_gui(self, trace=None, heat_map=None, p_bar=None):

        # Sanity checks
        if trace is not None and not isinstance(trace, TraceInterface):
            raise TypeError(
                'set_gui(): given trace instance does not inherit from TraceInterface'
            )
        if heat_map is not None and not isinstance(heat_map, HeatMapInterface):
            raise TypeError(
                'set_gui(): given heat_map instance does not inherit from HeatMapInterface'
            )
        if p_bar is not None and not isinstance(p_bar, PBarInterface):
            raise TypeError(
                'set_gui(): given p_bar instance does not inherit from PBarInterface'
            )

        # Save references in internal variables
        # [None if not given]
        self._1d_trace = trace
        self._2d_heat_map = heat_map
        self._p_bar = p_bar

        # Set axes labels
        if self._1d_trace is not None:
            self._1d_trace.set_lbls(
                x_str='tau [s]',
                y_str='State'
            )

        if self._2d_heat_map is not None:
            self._2d_heat_map.set_lbls(
                x_str='tau [s]',
                y_str='Repetition'
            )

        return 0

    def set_params(self, freq, pwr,
                   tau_start, tau_stop, n_pts,
                   n_soft_reps, n_hrdw_reps,
                   aom_delay=500e-9):

        self._freq = freq
        self._pwr = pwr
        self._tau_start = tau_start
        self._tau_stop = tau_stop
        self._n_pts = n_pts
        self._n_soft_reps = n_soft_reps
        self._n_hrdw_reps = n_hrdw_reps
        self._aom_delay = aom_delay

        # Allocate data arrays
        self._alloc_ars()

        return 0

    def run(self, clear_data=False, rfcs=True, rfcs_dur=12):

        try:

            self._setup_hardware()

            # Remove data from previous runs
            if clear_data:
                self.clear_data()

            self._is_running = True

            # Main measurement iteration loop
            while self._is_running and self._soft_rep_idx < self._n_soft_reps:

                # Run actual iteration
                self._make_iter()

                # Increment _soft_rep_idx only after the iteration is complete
                # (if iteration is interrupted before completion it should
                # not count as done)
                self._soft_rep_idx += 1

                # Update GUI
                self._update_gui()

                if rfcs:
                    self.rfcs(rfcs_dur=rfcs_dur)

            # Measurement is complete or paused
            self._hardware_off()
            self._is_running = False
            # [The main difference from interruption due to exception is that
            # here the loop waits for the iteration in progress to complete
            # before switching hardware off]

            return 0

        # Any other error
        except Exception as exc_obj:
            self.interrupt()
            raise exc_obj

    def pause(self):

        # The main difference between pause and interruption due to exception
        # is that, after calling pause(), the loop waits for the iteration in
        # progress to complete.

        self._is_running = False
        return 0

    def interrupt(self):
        self._hardware_off()
        self._is_running = False
        return 0

    def clear_data(self):

        # Reset sweep index
        self._soft_rep_idx = 0

        # Re-allocate data arrays
        self._alloc_ars()

        return 0

    # Technical methods

    def _alloc_ars(self):
        """Allocate/re-allocate data arrays

        :return: (int) status code: 0 - Ok
                            Exception - Error
        """

        # Delete existing arrays
        if self.tau_ar is not None:
            del self.tau_ar

        if self.avg_ar is not None:
            del self.avg_ar

        if self.raw_ar is not None:
            del self.raw_ar

        # Allocate empty arrays
        self.tau_ar = np.linspace(
            start=self._tau_start,
            stop=self._tau_stop,
            num=self._n_pts
        )
        self.avg_ar = np.zeros(
            shape=self._n_pts,
            dtype=float
        )
        self.raw_ar = np.zeros(
            shape=(self._n_soft_reps, self._n_pts),
            dtype=float
        )

        return 0

    def _setup_hardware(self):

        # Activate interface: "logic-reset" - not full hardware reset,
        # but reset of all the params to which any logic has access
        # (including params used by any other interface to this device)
        # and configuring the device to operate for this interface
        self._mw_src.activate_interface()
        self._gated_ctr.activate_interface()
        self._p_gen.activate_interface()

        # Configure mw_src
        self._mw_src.set_pwr(pwr=self._pwr)
        self._mw_src.set_freq(freq=self._freq)
        self._mw_src.on()

        # Init gated_ctr
        self._gated_ctr.init_ctr(
            bin_number=2 * self._n_pts * self._n_hrdw_reps,
            gate_type='RF'
        )

        # Configure p_gen
        self._p_gen.write(
            pb_obj=self._gen_p_block()
        )
        self._p_gen.set_rep(
            rep_num=self._n_hrdw_reps
        )

        return 0

    def _gen_p_block(self, safety_window=200e-9):

        # ========== Prepare individual blocks =========

        # ------- pre_init_pb --------
        # Long optical pumping before first pulse

        pre_init_pb = pb.PulseBlock(
            p_obj_list=[
                po.PTrue(ch='aom', dur=3e-6)
            ]
        )

        # ------- all_off_pb --------
        # Switch all channels off after the full pulse sequence

        all_off_pb = pb.PulseBlock(
            p_obj_list=[
                po.PFalse(ch='aom', dur=1e-6),
                po.PFalse(ch='mw_gate', dur=1e-6),
                po.PFalse(ch='ctr_gate', dur=1e-6)
            ]
        )

        # ========== Construct full rabi_pb =========

        rabi_pb = pb.PulseBlock(name='RabiPB')

        # Pre-init block
        rabi_pb.insert_pb(pb_obj=pre_init_pb)

        # rabi_element ** n_pts
        for tau in self.tau_ar:
            rabi_pb.append_pb(
                pb_obj=self._gen_rabi_elem(
                    tau=tau,
                    safety_window=safety_window
                )
            )

        # All-off block
        rabi_pb.append_pb(pb_obj=all_off_pb)

        # Default values
        rabi_pb.dflt_dict = dict(
            aom=po.DFalse(),
            ctr_gate=po.DFalse(),
            mw_gate=po.DFalse()
        )

        # Correct for aom delay
        rabi_pb.add_offset(
            offset_dict=dict(
                aom=-self._aom_delay
            )
        )

        # Duration of full RabiPB
        # [for gated_ctr.get_count_ar() timeout]
        self._rabi_pb_dur = rabi_pb.dur

        return rabi_pb

    @staticmethod
    def _gen_rabi_elem(tau=0, safety_window=200e-9):
        rabi_elem = pb.PulseBlock()

        # Init green_aom pulse
        rabi_elem.insert(
            p_obj=po.PTrue(
                ch='aom',
                dur=2e-6 + 2*safety_window
            )
        )

        # Normalization pulse
        rabi_elem.insert(
            p_obj=po.PTrue(
                ch='ctr_gate',
                t0=1.5e-6,
                dur=0.5e-6
            )
        )

        # mw pulse
        rabi_elem.append(
            p_obj=po.PTrue(
                ch='mw_gate',
                dur=tau,
                t0=4 * safety_window
            )
        )

        tmp_dur = rabi_elem.dur + 4*safety_window

        # Readout AOM pulse
        rabi_elem.insert(
            p_obj=po.PTrue(
                ch='aom',
                dur=1e-6,
                t0=tmp_dur
            )
        )

        # readout ctr_gate pulse
        rabi_elem.insert(
            p_obj=po.PTrue(
                ch='ctr_gate',
                dur=0.5e-6,
                t0=tmp_dur
            )
        )

        return rabi_elem

    def _make_iter(self):

        self._gated_ctr.start_counting()
        self._p_gen.start()

        # Read data from counter
        timeout = 10 * self._rabi_pb_dur * self._n_hrdw_reps
        flat_cnt_ar = self._gated_ctr.get_count_ar(timeout=timeout)

        # Sum counts from different hardware repetitions
        cnt_ar = np.reshape(
            a=flat_cnt_ar,
            newshape=(self._n_hrdw_reps, 2 * self._n_pts)
        )
        cnt_ar = np.sum(a=cnt_ar, axis=0)

        # Calculate state array
        state_ar = np.divide(
            cnt_ar[1::2],  # readout counts
            cnt_ar[::2]    # normalization counts
        )

        # Update data arrays
        self.raw_ar[self._soft_rep_idx][:] = state_ar
        self.avg_ar[:] = self.raw_ar[:self._soft_rep_idx + 1].mean(axis=0)

        self._p_gen.stop()

    def rfcs(self, rfcs_dur=12):
        self._p_gen.stop()
        self._p_gen.write(
            pb_obj=self._gen_rfcs_block()
        )
        self._p_gen.set_rep(rep_num=0)
        self._p_gen.start()

        self._rfcs.rfcs()
        time.sleep(rfcs_dur)

        self._p_gen.stop()
        self._p_gen.write(
            pb_obj=self._gen_p_block()
        )
        self._p_gen.set_rep(
            rep_num=self._n_hrdw_reps
        )

        return 0

    @staticmethod
    def _gen_rfcs_block():

        optimize_pb = pb.PulseBlock()
        optimize_pb.insert(
            p_obj=po.PTrue(ch='aom', dur=1e-6)
        )
        optimize_pb.insert(
            p_obj=po.PTrue(ch='mw_gate', dur=1e-6)
        )
        optimize_pb.dflt_dict = dict(
            aom=po.DTrue(),
            mw_gate=po.DTrue()
        )
        return optimize_pb

    def _update_gui(self):

        if self._1d_trace is not None:
            self._1d_trace.set_data(
                x_ar=self.tau_ar,
                y_ar=self.avg_ar
            )

        if self._2d_heat_map is not None:
            self._2d_heat_map.set_data(
                x_ar=self.tau_ar,
                y_ar=np.arange(start=1, stop=self._soft_rep_idx + 1),
                z_ar=self.raw_ar[:self._soft_rep_idx]
            )

        if self._p_bar is not None:
            self._p_bar.set_value(
                value=100 * self._soft_rep_idx / self._n_soft_reps
            )

    def _hardware_off(self):
        self._mw_src.off()
        self._p_gen.stop()
        self._gated_ctr.terminate_counting()



"""Wrapper for NI654x driver to implement SimplePGenInterface"""

from pylabnet.hardware.interface.simple_p_gen import SimplePGenInterface
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase
import pickle

# TODO: add automatic frequency setting
# TODO: (analyze given PulseBlock, find smallest interval, set 1/10 sampling rate)


class Wrap(SimplePGenInterface):

    def __init__(self, ni654x_inst):
        self._dev = ni654x_inst

    def activate_interface(self):

        # TODO: add mutex check and lock (force lock option)

        # Reset device
        #   store hardware settings which are not controlled by logic,
        #   to restore them after reset()
        #   [logic does not know anything about this params, so it should not
        #   introduce any changes to them by calling activate_interface()].
        tmp_map_dict = self._dev.map_dict
        tmp_samp_rate = self._dev.get_samp_rate()
        #   full hardware reset
        self._dev.reset()

        # Restore hardware settings which are not controlled by logic
        # but were changed by self._dev.reset()
        self._dev.map_dict = tmp_map_dict
        self._dev.set_samp_rate(samp_rate=tmp_samp_rate)

        # Configure the device to operate as Simple Pulse Generator
        self._dev.set_mode(mode_string='W')
        self._dev.set_active_chs(chs_str='0-31')

        return 0

    def write(self, pb_obj, len_adj=True):

        self._dev.clr_mem()

        return self._dev.write_wfm(
            pb_obj=pb_obj,
            len_adj=len_adj
        )

    def set_rep(self, rep_num):
        return self._dev.set_rep(
            rep_num=rep_num
        )

    def start(self):
        return self._dev.start()

    def stop(self):
        return self._dev.stop()

    def get_status(self):
        """Get status of the device

        0 - 'Idle'
        1 - 'Running'
        Exception is produced in the case of any error
        (for example, connection to the device is lost)

        :return: (int) status code
                 Exception is produced in the case of error
        """

        return self._dev.get_status()


class Service(ServiceBase):

    def exposed_activate_interface(self):
        return self._module.activate_interface()

    def exposed_write(self, pb_obj_pckl, len_adj=True):

        pb_obj = pickle.loads(pb_obj_pckl)

        return self._module.write(
            pb_obj=pb_obj,
            len_adj=len_adj
        )

    def exposed_set_rep(self, rep_num):
        return self._module.set_rep(
            rep_num=rep_num
        )

    def exposed_start(self):
        return self._module.start()

    def exposed_stop(self):
        return self._module.stop()

    def exposed_get_status(self):
        return self._module.get_status()


class Client(ClientBase, SimplePGenInterface):

    def activate_interface(self):
        return self._service.exposed_activate_interface()

    def write(self, pb_obj, len_adj=True):

        pb_obj_pckl = pickle.dumps(pb_obj)

        return self._service.exposed_write(
            pb_obj_pckl=pb_obj_pckl,
            len_adj=len_adj
        )

    def set_rep(self, rep_num):
        return self._service.exposed_set_rep(
            rep_num=rep_num
        )

    def start(self):
        return self._service.exposed_start()

    def stop(self):
        return self._service.exposed_stop()

    def get_status(self):
        return self._service.exposed_get_status()



# -*- coding: utf-8 -*-

"""This file contains the pylabnet client which allows qudi to access
Swabian Instruments TT through pylabnet network as SlowCounter.

Steps:
- instantiate Time Tagger
- instantiate pylabnet-SlowCtrWrap (pass ref to Time Tagger as tagger)
- instantiate pylabnet-SlowCtrService and assign module to the created wrapper
- start pylabnet-server for SlowCtrService
- in qudi, instantiate SlowCtrClient as one of the hardware modules

Currently broken, as the link from pylabnet to Qudi has been severed
"""

# from network.core.module import Base, ConfigOption
# from interface.slow_counter_interface import SlowCounterInterface, SlowCounterConstraints, CountingMode
# import rpyc
# import pickle


# class SlowCtrClient(Base, SlowCounterInterface):

#     _modclass = 'SlowCounterClient'
#     _modtype = 'hardware'

#     _cfg_host = ConfigOption(name='host', missing='error')
#     _cfg_port = ConfigOption(name='port', missing='error')

#     def __init__(self, config, **kwargs):
#         super().__init__(config=config, **kwargs)

#         self._host = self._cfg_host
#         self._port = self._cfg_port

#         # References to TT.Counter measurement
#         self._connection = None
#         self._service = None

#     def on_activate(self):

#         self._connection = rpyc.connect(
#             host=self._host,
#             port=self._port,
#             config={'allow_public_attrs': True}
#         )

#         self._service = self._connection.root
#         self.log.info('on_activate(): connected to SlowCounterService: host={} port={}'
#                       ''.format(self._host, self._port))

#     def on_deactivate(self):
#         self._connection.close()
#         self.log.info('on_deactivate(): closed connection to SlowCounterService: host={} port={}'
#                       ''.format(self._host, self._port))

#     def set_up_clock(self, clock_frequency=None, clock_channel=None):
#         """
#         Sets sample clock frequency for the Counter measurement.


#         :param clock_frequency: (float) sample clock frequency. If not given,
#                                 configuration value is used
#         :param clock_channel: ignored (internal timebase is used to generate
#                               sample clock signal)

#         :return: (int) operation status code: 0 - OK
#                                              -1 - Error
#         """

#         return self._service.exposed_set_up_clock(
#             clock_frequency=clock_frequency,
#             clock_channel=clock_channel
#         )

#     def set_up_counter(self,
#                        counter_channels=None,
#                        sources=None,
#                        clock_channel=None,
#                        counter_buffer=None):
#         """
#         Configures the actual counter with a given clock.


#          (list of int) [optional] list of channels
#                      to count clicks on. If not given, config value is used.

#         :param counter_buffer: (int) [optional] size of the memory buffer.
#                                If not given, config value is used.

#         :param counter_channels: ignored
#             This argument should not be used. Counter GUI initializes set of plot curves
#             self.curves during its on_activate() method. It basically calls
#             counter_hardware.get_counter_channels() and uses this list to init self.curves
#             Only after that user can click "Start" button, which will call set_up_counter().
#             And since GUI already has inited set of curves, set of channels must not be
#             modified here! It will case GUI to fail.

#         :param sources: ignored
#         :param clock_channel: ignored

#         :return: (int) operation status code: 0 - OK
#                                              -1 - Error
#         """

#         return self._service.exposed_set_up_counter(
#             counter_channels=counter_channels,
#             sources=sources,
#             clock_channel=clock_channel,
#             counter_buffer=counter_buffer
#         )

#     def close_clock(self):
#         """
#         Closes the clock.

#         :return: (int) error code: 0 - OK
#                                   -1 - Error
#         """

#         return self._service.exposed_close_clock()

#     def close_counter(self):
#         """
#         Closes the counter and cleans up afterwards.

#         :return: (int) error code: 0 - OK
#                                   -1 - Error
#         """

#         return self._service.exposed_close_counter()

#     def get_counter(self, samples=1):
#         """
#         Returns the current counts per second of the counter.

#         :param samples: (int) [optional] number of samples to read in one go
#                         (default is one sample)

#         :return: numpy.array((samples, uint32), dtype=np.uint32)
#         array of count rate [counts/second] arrays of length samples for each click channel
#         Empty array [] is returned in the case of error.
#         """

#         res_pickle = self._service.exposed_get_counter(samples=samples)
#         return pickle.loads(res_pickle)

#     def get_constraints(self):
#         """
#         Retrieve the hardware constrains from the counter device.

#         :return: (SlowCounterConstraints) object with constraints for the counter
#         """

#         constraints = SlowCounterConstraints()
#         # TODO: check values
#         constraints.min_count_frequency = 1
#         constraints.max_count_frequency = 10e9
#         constraints.max_detectors = 8
#         constraints.counting_mode = [CountingMode.CONTINUOUS]

#         return constraints

#     def get_counter_channels(self):
#         """
#         Returns the list of click channel numbers.

#         :return: (list of int) list of click channel numbers
#         """

#         res_pickle = self._service.exposed_get_counter_channels()
#         return pickle.loads(res_pickle)



from pylabnet.hardware.interface.mw_src import MWSrcInterface
from pylabnet.hardware.interface.gated_ctr import GatedCtrInterface
from pylabnet.hardware.interface.simple_p_gen import SimplePGenInterface
from pylabnet.gui.output_interface import TraceInterface, HeatMapInterface, PBarInterface
import pylabnet.utils.pulseblock.pulse as po
import pylabnet.utils.pulseblock.pulse_block as pb
import numpy as np


class ODMR:
    def __init__(self):

        # Hardware instances
        self._mw_src = None
        self._gated_ctr = None
        self._p_gen = None

        # GUI instances
        self._1d_trace = None
        self._2d_heat_map = None
        self._p_bar = None

        # Parameters
        self._start_f = None
        self._stop_f = None
        self._n_pts = None
        self._n_swps = None
        self._mw_pwr = None
        self._accum_dur = 1e-3

        # Internal data arrays
        self.freq_ar = None  # 1D numpy float array
        self.avg_ar = None  # 1D numpy float array
        self.raw_ar = None  # 2D numpy float array

        # Current sweep index
        self._is_running = False
        self._swp_idx = 0

    # User interface methods

    def set_hardware(self, mw_src, gated_ctr, p_gen):

        # Sanity checks
        if not isinstance(mw_src, MWSrcInterface):
            raise TypeError(
                'set_hardware(): given mw_src instance does not inherit from MWSrcInterface'
            )
        if not isinstance(gated_ctr, GatedCtrInterface):
            raise TypeError(
                'set_hardware(): given gated_ctr instance does not inherit from GatedCtrInterface'
            )
        if not isinstance(p_gen, SimplePGenInterface):
            raise TypeError(
                'set_hardware(): given p_gen instance does not inherit from SimplePGenInterface'
            )

        # Save references in internal variables
        self._mw_src = mw_src
        self._gated_ctr = gated_ctr
        self._p_gen = p_gen

        return 0

    def set_gui(self, trace=None, heat_map=None, p_bar=None):

        # Sanity checks
        if trace is not None and not isinstance(trace, TraceInterface):
            raise TypeError(
                'set_gui(): given trace instance does not inherit from TraceInterface'
            )
        if heat_map is not None and not isinstance(heat_map, HeatMapInterface):
            raise TypeError(
                'set_gui(): given heat_map instance does not inherit from HeatMapInterface'
            )
        if p_bar is not None and not isinstance(p_bar, PBarInterface):
            raise TypeError(
                'set_gui(): given p_bar instance does not inherit from PBarInterface'
            )

        # Save references in internal variables
        # [None if not given]
        self._1d_trace = trace
        self._2d_heat_map = heat_map
        self._p_bar = p_bar

        # Set axes labels
        if self._1d_trace is not None:
            self._1d_trace.set_lbls(
                x_str='Frequency [Hz]',
                y_str='Count rate [Cts/s]'
            )

        if self._2d_heat_map is not None:
            self._2d_heat_map.set_lbls(
                x_str='Frequency [Hz]',
                y_str='Repetition'
            )

        return 0

    def set_params(self, start_f, stop_f, n_pts, n_swps, mw_pwr, accum_dur=1e-3):
        # Save params to internal variables
        self._start_f = start_f
        self._stop_f = stop_f
        self._n_pts = n_pts
        self._n_swps = n_swps
        self._mw_pwr = mw_pwr
        self._accum_dur = accum_dur

        # Allocate data arrays
        self._alloc_ars()

        return 0

    def run(self, clear_data=False):

        try:

            self._setup_hardware()

            # Remove data from previous runs
            if clear_data:
                self.clear_data()

            self._is_running = True

            # Main measurement iteration loop
            while self._is_running and self._swp_idx < self._n_swps:

                # Run actual iteration
                self._make_iter()

                # Increment _swp_idx only after the iteration is complete
                # (if iteration is interrupted before completion it should
                # not count as done)
                self._swp_idx += 1

                # Update GUI
                self._update_gui()

            # Measurement is complete or paused
            self._hardware_off()
            self._is_running = False
            # [The main difference from interruption due to exception is that
            # here the loop waits for the iteration in progress to complete
            # before switching hardware off]

        except KeyboardInterrupt:
            # Measurement is interrupted by user
            return self.interrupt()

        except Exception as exc_obj:
            # Any other error
            self.interrupt()
            raise exc_obj

    def run_in_bkgr(self):
        # TODO: implement
        raise NotImplementedError(
            'Running in a background thread is not implemented'
        )

    def pause(self):

        # The main difference between pause and interruption due to exception
        # is that, after calling pause(), the loop waits for the iteration in
        # progress to complete.

        self._is_running = False
        return 0

    def interrupt(self):
        self._hardware_off()
        self._is_running = False
        return 0

    def clear_data(self):

        # Reset sweep index
        self._swp_idx = 0

        # Re-allocate data arrays
        self._alloc_ars()

        return 0

    # Technical methods

    def _alloc_ars(self):
        """Allocate/re-allocate data arrays

        :return: (int) status code: 0 - Ok
                            Exception - Error
        """

        # Delete existing arrays
        if self.freq_ar is not None:
            del self.freq_ar

        if self.avg_ar is not None:
            del self.avg_ar

        if self.raw_ar is not None:
            del self.raw_ar

        # Allocate empty arrays
        self.freq_ar = np.linspace(
            start=self._start_f,
            stop=self._stop_f,
            num=self._n_pts
        )
        self.avg_ar = np.zeros(
            shape=self._n_pts,
            dtype=float
        )
        self.raw_ar = np.zeros(
            shape=(self._n_swps, self._n_pts),
            dtype=float
        )

        return 0

    def _setup_hardware(self):

        # Activate interface: "logic-reset" - not full hardware reset,
        # but reset of all the params to which any logic has access
        # (including params used by any other interface to this device)
        # and configuring the device to operate for this interface
        self._mw_src.activate_interface()
        self._gated_ctr.activate_interface()
        self._p_gen.activate_interface()

        # Configure mw_src to freq sweep
        self._mw_src.set_pwr(pwr=self._mw_pwr)
        self._mw_src.set_freq_swp(
            start=self._start_f,
            stop=self._stop_f,
            n_pts=self._n_pts
        )
        self._mw_src.on()

        # Init gated_ctr
        self._gated_ctr.init_ctr(
            bin_number=self._n_pts,
            gate_type='RF'
        )

        # Configure p_gen
        self._p_gen.write(
            pb_obj=self._gen_p_block()
        )
        self._p_gen.set_rep(rep_num=1)

        return 0

    def _gen_p_block(self, debug=False):

        # Technical params

        # safety window between mw_src makes a step and
        # before counter starts accumulation of counts
        safety_window = 0.1 * self._accum_dur

        # Duration of sweep-step pulse
        step_pulse_dur = 0.1 * self._accum_dur  # 200e-6

        # ========== Prepare individual blocks =========

        # ------- first_pb --------
        # First point in the sweep:
        #   mw_src should reset sweep position to start_f,
        #   so there is no need to send step_trig pulse

        first_pb = pb.PulseBlock()

        # long safety window in the beginning
        first_pb.insert(
            p_obj=po.PFalse(
                ch='ctr_gate',
                dur=self._accum_dur
            )
        )

        # first ctr_gate pulse
        first_pb.append(
            p_obj=po.PTrue(
                ch='ctr_gate',
                dur=self._accum_dur
            )
        )

        # ------- step_pb --------
        # This pb is repeated for every subsequent sweep step:
        #   send ctr_trig pulse, wait for a safety window and
        #   start accumulating pulses

        step_pb = pb.PulseBlock()

        # step pulse to mw_src
        step_pb.insert(
            p_obj=po.PTrue(
                ch='mw_step',
                dur=step_pulse_dur
            )
        )

        # ctr_gate pulse
        step_pb.append(
            p_obj=po.PTrue(
                ch='ctr_gate',
                dur=self._accum_dur,
                t0=safety_window
            )
        )

        # ------- off_pb --------
        # Once sweep is complete, set all outputs to low

        off_pb = pb.PulseBlock()

        off_pb.insert(
            p_obj=po.PFalse(
                ch='aom',
                dur=safety_window
            )
        )
        off_pb.insert(
            p_obj=po.PFalse(
                ch='mw_gate',
                dur=safety_window
            )
        )
        off_pb.insert(
            p_obj=po.PFalse(
                ch='mw_step',
                dur=safety_window
            )
        )
        off_pb.insert(
            p_obj=po.PFalse(
                ch='ctr_gate',
                dur=safety_window
            )
        )

        # ========== Construct full odmr_pb =========

        odmr_pb = pb.PulseBlock(name='ODMRPulseBlock')

        # Default values
        odmr_pb.dflt_dict = dict(
            aom=po.DFalse(),
            ctr_gate=po.DFalse(),
            mw_gate=po.DFalse(),
            mw_step=po.DFalse(),
        )

        # first_bp
        odmr_pb.insert_pb(pb_obj=first_pb)

        # append step_bp ** (_n_pts-1)
        for _ in np.arange(start=1, stop=self._n_pts):
            odmr_pb.append_pb(pb_obj=step_pb)

        # aom and mw_gate are High during the whole block
        tmp_dur = odmr_pb.dur
        odmr_pb.insert(
            p_obj=po.PTrue(ch='aom', dur=tmp_dur)
        )
        odmr_pb.insert(
            p_obj=po.PTrue(ch='mw_gate', dur=tmp_dur)
        )

        # append off_pb
        odmr_pb.append_pb(pb_obj=off_pb)

        if debug:
            return dict(
                first_pb=first_pb,
                step_pb=step_pb,
                off_pb=off_pb,
                odmr_pb=odmr_pb
            )
        else:
            return odmr_pb

    def _make_iter(self):

        self._gated_ctr.start_counting()
        self._mw_src.reset_swp_pos()
        self._p_gen.start()

        # Read data from counter
        timeout = 10 * self._accum_dur * self._n_pts
        # read and convert to numpy float array
        raw_cnt_ar = np.asarray(
            a=self._gated_ctr.get_count_ar(timeout=timeout),
            dtype=float
        )
        # calculate count rate
        cnt_rate_ar = raw_cnt_ar / self._accum_dur

        # Update data arrays
        self.raw_ar[self._swp_idx][:] = cnt_rate_ar
        self.avg_ar[:] = self.raw_ar[:self._swp_idx+1].mean(axis=0)

        self._p_gen.stop()

    def _update_gui(self):

        if self._1d_trace is not None:
            self._1d_trace.set_data(
                x_ar=self.freq_ar,
                y_ar=self.avg_ar
            )

        if self._2d_heat_map is not None:
            self._2d_heat_map.set_data(
                x_ar=self.freq_ar,
                y_ar=np.arange(start=1, stop=self._swp_idx + 1),
                z_ar=self.raw_ar[:self._swp_idx]
            )

        if self._p_bar is not None:
            self._p_bar.set_value(
                value=100 * self._swp_idx / self._n_swps
            )

    def _hardware_off(self):
        self._mw_src.off()
        self._p_gen.stop()
        self._gated_ctr.terminate_counting()



""" Hardware wrapper for Swabian Instruments TimeTagger, including client and server access classes """

import TimeTagger as TT
import pickle
import socket
from pylabnet.utils.logging.logger import LogHandler


class Wrap:

    def __init__(self, tagger, ch_list=[1], logger=None):
        """Instantiate count monitor

        :param tagger: instance of TimeTagger class
        :param ch_list: list of channels to count
        :param logger: instance of LogClient class, optional
        """

        # Log
        self.log = LogHandler(logger=logger)

        # Reference to tagger
        self._tagger = tagger

        # Log device ID information to demonstrate that connection indeed works
        serial = self._tagger.getSerial()
        model = self._tagger.getModel()
        self.log.info(
            'Got reference to Swabian Instruments TimeTagger device \n'
            'Serial number: {0}, Model: {1}'
            ''.format(serial, model)
        )

        # Counter instance
        self._ctr = None

        # Set click channels
        self._ch_list = ch_list
        self.set_ch_assignment(ch_list=ch_list)

    def start_ctr(self, bin_width=1000000000, n_bins=10000):
        """Start counter

        :param bin_width: integer in ps for width of count bins
        :param n_bins: integer number of bins to store before
                            wrapping around
        """
        
        # Instantiate Counter instance, see TT documentation
        self._ctr = TT.Counter(
            self._tagger, 
            channels=self._ch_list, 
            binwidth=bin_width, 
            n_values=n_bins
        )

    def clear_ctr(self):
        """Resets the array to zero and restarts the measurement.
            See the clear() method of Counter class in TT
        """

        # Clear counter (see TT documentation)
        self._ctr.clear()

    def get_counts(self):
        """Gets a 2D array of counts on all channels. See the 
            getData() method of Counter class in TT
        """

        # Get count data (see TT documentation)
        return self._ctr.getData()

    def get_x_axis(self):
        """Gets the x axis in picoseconds for the count array.
            See the getIndex() method of Counter class in TT
        """

        # Get x axis
        return self._ctr.getIndex()

    def set_ch_assignment(self, ch_list=[1]):
        """Sets the ch_list attribute of the wrapper to a valid
            list of channel numbers as desired by TT, also 
            configures the default naming convention for channels

        :param ch_list: list of integer numbers of channels,
                        following the convention 1 ... 8
                        for rising edge and negative for falling
        """

        # Set attribute to validated ch_list
        self._ch_list = ch_list



import rpyc
import threading


class GenericServer:
    def __init__(self, service, port, host='localhost'):

        self._server = rpyc.ThreadedServer(
            service=service,
            hostname=host,
            port=port,
            protocol_config={
                'allow_public_attrs': True,
                'sync_request_timeout': 300
            }
        )

        self._server_thread = threading.Thread(
            target=self._start_server,
            args=(self._server,)
        )

    def start(self):
        self._server_thread.start()

    def stop(self):
        pass

    @staticmethod
    def _start_server(server_obj):
        server_obj.start()




'''
This module contains utility functions used for pylabnet decorators.
'''


def get_signature(*args, **kwargs):
    """ Gets printable function signature"""
    args_repr = [repr(a) for a in args]
    kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
    signature = ", ".join(args_repr + kwargs_repr)
    return signature



import numpy as np


class PID:
    """Generic class for PID locking"""

    def __init__(self, p=0, i=0, d=0, setpoint=0, memory=20):
        """ Constructor for PID class
        
        :rtype: object
        :param p: proportional gain
        :param i: integral
        :param d: differential
        :param setpoint: setpoint for process variable
        :param memory: number of samples for integral memory
        """

        self.p = p
        self.i = i
        self.d = d
        self.memory = memory
        self.setpoint = setpoint
        self._pv = np.zeros(self.memory)
        self.cv = 0
        self.error = 0

    def set_parameters(self, p=None, i=None, d=None, setpoint=None, memory=None):
        """ Sets parameters of PID controller

        :param p: proportional gain
        :param i: integral
        :param d: differential
        :param setpoint: setpoint for process variable
        :param memory: number of samples for integral memory
        """

        if p is not None:
            self.p = p
        if i is not None:
            self.i = i
        if d is not None:
            self.d = d
        if memory is not None:
            self.memory = memory
        if setpoint is not None:
            self.setpoint = setpoint

        # Update process variable history to have appropriate length if too short
        pv_length = len(self._pv)
        if pv_length < self.memory:

            # Pad constants onto the beginning of the array
            self._pv = np.hstack((np.ones(self.memory-pv_length)*self._pv[0], self._pv))

    def set_pv(self, pv=np.zeros(10)):
        """ Sets process variable

        :param pv: process variable (measured value of process to be locked).
            This should ideally be a numpy array of recent data points
        """

        # Check the length of the input
        pv_length = len(pv)

        # If it's too short, append it onto the current pv
        if pv_length < self.memory:
            self._pv = np.append(self._pv[self.memory-pv_length:], pv)

        # Otherwise just take the last elements
        else:
            self._pv = pv[pv_length-self.memory:]

    def set_cv(self):
        """Calculates the appropriate value of the control variable"""

        # Calculate error
        error = self._pv - self.setpoint
        self.error = error[-1]

        # Calculate response
        self.cv = self.p*error[-1] + self.i*np.sum(error)/self.memory + self.d*(error[-1] - error[-2])



""" Launches the wavemeter monitor/control application """

from pylabnet.launchers.launcher import Launcher
from pylabnet.launchers.servers import high_finesse_ws7, nidaqmx_wlm
from pylabnet.scripts.wavemeter import wlm_monitor


def main():

    launcher = Launcher(
        script=[wlm_monitor],
        server_req=[high_finesse_ws7, nidaqmx_wlm],
        gui_req=['wavemeter_monitor'],
        params=[None]
    )
    launcher.launch()


if __name__ == '__main__':
    main()



import socket

from pylabnet.hardware.wavemeter.high_finesse_ws7 import Driver
from pylabnet.network.client_server.high_finesse_ws7 import Service, Client
from pylabnet.network.core.generic_server import GenericServer


def launch(**kwargs):
    """ Connects to HF WS7 Wavemeter and launches server

    :param kwargs: (dict) containing relevant kwargs
        :logger: instance of LogClient for logging purposes
        :port: (int) port number for the Cnt Monitor server
    """

    # Instantiate Logger
    wavemeter_logger = kwargs['logger']

    # Instantiate Wavemeter object
    hf_wlm = Driver(logger=wavemeter_logger)

    # Instantiate Server
    wavemeter_service = Service()
    wavemeter_service.assign_module(module=hf_wlm)
    wavemeter_service.assign_logger(logger=wavemeter_logger)
    wavemeter_server = GenericServer(
        service=wavemeter_service,
        host=socket.gethostbyname(socket.gethostname()),
        port=kwargs['port']
    )

    wavemeter_server.start()



import rpyc
from pylabnet.utils.logging.logger import LogHandler


class ServiceBase(rpyc.Service):

    _module = None
    log = LogHandler()

    def on_connect(self, conn):
        # code that runs when a connection is created
        # (to init the service, if needed)
        self.log.info('Client connected')

    def on_disconnect(self, conn):
        # code that runs after the connection has already closed
        # (to finalize the service, if needed)
        self.log.info('Client disconnected')

    def assign_module(self, module):
        self._module = module

    def assign_logger(self, logger=None):
        self.log = LogHandler(logger=logger)



""" Launches the continuous count monitor application """

from pylabnet.launchers.launcher import Launcher
from pylabnet.launchers.servers import si_tt_cnt_monitor
from pylabnet.scripts.counter import monitor_counts


def main():

    launcher = Launcher(
        script=[monitor_counts],
        server_req=[si_tt_cnt_monitor],
        gui_req=['count_monitor'],
        params=[None]
    )
    launcher.launch()


if __name__ == '__main__':
    main()



from pylabnet.hardware.interface.gated_ctr import GatedCtrInterface
from pylabnet.hardware.interface.simple_p_gen import SimplePGenInterface
from pylabnet.gui.output_interface import TraceInterface, HeatMapInterface, PBarInterface
import pylabnet.utils.pulseblock.pulse as po
import pylabnet.utilspulseblock.pulse_block as pb
import numpy as np
import time


class CountTrace:
    def __init__(self):

        # Hardware instances
        self._gated_ctr = None
        self._p_gen = None
        self._rfcs = None

        # GUI instances
        self._1d_trace = None
        self._2d_heat_map = None
        self._p_bar = None

        # Parameters
        self._laser_dur = None
        self._l_margin = None
        self._r_margin = None
        self._n_pts = None
        self._relax = None
        self._hrdw_reps = None
        self._soft_reps = None

        # Internal data arrays
        self.t_ar = None  # 1D numpy float array
        self.sum_ar = None  # 1D numpy float array
        self.raw_ar = None  # 2D numpy float array

        # Current sweep index
        self._is_running = False
        self._soft_rep_idx = 0

        # Approximate duration of full pulse block
        # [for gated_ctr.get_count_ar() timeout]
        self._pb_dur = -1

    # User interface methods

    def set_hardware(self, gated_ctr, p_gen, rfcs=None):

        # Sanity checks
        if not isinstance(gated_ctr, GatedCtrInterface):
            raise TypeError(
                'set_hardware(): given gated_ctr instance does not inherit from GatedCtrInterface'
            )
        if not isinstance(p_gen, SimplePGenInterface):
            raise TypeError(
                'set_hardware(): given p_gen instance does not inherit from SimplePGenInterface'
            )

        # Save references in internal variables
        self._gated_ctr = gated_ctr
        self._p_gen = p_gen

        if rfcs is not None:
            self._rfcs = rfcs

        return 0

    def set_gui(self, trace=None, heat_map=None, p_bar=None):

        # Sanity checks
        if trace is not None and not isinstance(trace, TraceInterface):
            raise TypeError(
                'set_gui(): given trace instance does not inherit from TraceInterface'
            )
        if heat_map is not None and not isinstance(heat_map, HeatMapInterface):
            raise TypeError(
                'set_gui(): given heat_map instance does not inherit from HeatMapInterface'
            )
        if p_bar is not None and not isinstance(p_bar, PBarInterface):
            raise TypeError(
                'set_gui(): given p_bar instance does not inherit from PBarInterface'
            )

        # Save references in internal variables
        # [None if not given]
        self._1d_trace = trace
        self._2d_heat_map = heat_map
        self._p_bar = p_bar

        # Set axes labels
        if self._1d_trace is not None:
            self._1d_trace.set_lbls(
                x_str='Time [s]',
                y_str='Counts'
            )

        if self._2d_heat_map is not None:
            self._2d_heat_map.set_lbls(
                x_str='Time [s]',
                y_str='Repetition'
            )

        return 0

    def set_params(self, laser_dur, l_margin, r_margin, n_pts, relax, hrdw_reps, soft_reps):
        # Save params to internal variables
        self._laser_dur = laser_dur
        self._l_margin = l_margin
        self._r_margin = r_margin
        self._n_pts = n_pts
        self._relax = relax
        self._hrdw_reps = hrdw_reps
        self._soft_reps = soft_reps

        # Allocate data arrays
        self._alloc_ars()

        return 0

    def run(self, clear_data=False, rfcs=True, rfcs_dur=12, rfcs_every=1):

        try:

            self._setup_hardware()

            # Remove data from previous runs
            if clear_data:
                self.clear_data()

            self._is_running = True

            # Main measurement iteration loop
            while self._is_running and self._soft_rep_idx < self._soft_reps:

                # Run actual iteration
                self._make_iter()

                # Increment _soft_rep_idx only after the iteration is complete
                # (if iteration is interrupted before completion it should
                # not count as done)
                self._soft_rep_idx += 1

                # Update GUI
                self._update_gui()

                if rfcs and self._soft_rep_idx % rfcs_every == 0:
                    self.rfcs(rfcs_dur=rfcs_dur)

            # Measurement is complete or paused
            self._hardware_off()
            self._is_running = False
            # [The main difference from interruption due to exception is that
            # here the loop waits for the iteration in progress to complete
            # before switching hardware off]

            return 0

        # Any other error
        except Exception as exc_obj:
            self.interrupt()
            raise exc_obj

    def pause(self):

        # The main difference between pause and interruption due to exception
        # is that, after calling pause(), the loop waits for the iteration in
        # progress to complete.

        self._is_running = False
        return 0

    def interrupt(self):
        self._hardware_off()
        self._is_running = False
        return 0

    def clear_data(self):

        # Reset sweep index
        self._soft_rep_idx = 0

        # Re-allocate data arrays
        self._alloc_ars()

        return 0

    def rfcs(self, rfcs_dur=12):
        self._p_gen.stop()
        self._p_gen.write(
            pb_obj=self._gen_rfcs_block()
        )
        self._p_gen.set_rep(rep_num=0)
        self._p_gen.start()

        self._rfcs.rfcs()
        time.sleep(rfcs_dur)

        self._p_gen.stop()
        self._p_gen.write(
            pb_obj=self._gen_p_block()
        )
        self._p_gen.set_rep(
            rep_num=self._hrdw_reps
        )

        return 0

    # Technical methods

    @staticmethod
    def _gen_rfcs_block():

        optimize_pb = pb.PulseBlock(
            p_obj_list=[
                po.PTrue(ch='aom', dur=1e-6)
            ],
            dflt_dict=dict(aom=po.DTrue())
        )

        return optimize_pb

    def _alloc_ars(self):
        """Allocate/re-allocate data arrays

        :return: (int) status code: 0 - Ok
                            Exception - Error
        """

        # Delete existing arrays
        if self.t_ar is not None:
            del self.t_ar

        if self.sum_ar is not None:
            del self.sum_ar

        if self.raw_ar is not None:
            del self.raw_ar

        # Allocate empty arrays
        self.t_ar = np.linspace(
            start=-self._l_margin,
            stop=self._laser_dur+self._r_margin,
            num=self._n_pts
        )
        self.sum_ar = np.zeros(
            shape=self._n_pts,
            dtype=float
        )
        self.raw_ar = np.zeros(
            shape=(self._soft_reps, self._n_pts),
            dtype=float
        )

        return 0

    def _setup_hardware(self):

        # Activate interface: "logic-reset" - not full hardware reset,
        # but reset of all the params to which any logic has access
        # (including params used by any other interface to this device)
        # and configuring the device to operate for this interface
        self._gated_ctr.activate_interface()
        self._p_gen.activate_interface()

        # Init gated_ctr
        self._gated_ctr.init_ctr(
            bin_number=self._n_pts * self._hrdw_reps,
            gate_type='RR'
        )

        # Configure p_gen
        self._p_gen.write(
            pb_obj=self._gen_p_block()
        )
        self._p_gen.set_rep(
            rep_num=self._hrdw_reps
        )

        return 0

    def _gen_p_block(self):

        count_trace_pb = pb.PulseBlock()

        # Sample clock
        t_step = (self._l_margin + self._laser_dur + self._r_margin) / self._n_pts

        for idx in range(self._n_pts):
            count_trace_pb.append(
                p_obj=po.PTrue(
                    ch='ctr_gate',
                    dur=t_step / 2,
                    t0=t_step / 2
                )
            )

        # Laser pulse
        count_trace_pb.insert(
            p_obj=po.PTrue(
                ch='aom',
                dur=self._laser_dur,
                t0=self._l_margin
            )
        )

        # Relaxation period and all off pulse block
        count_trace_pb.append(
            p_obj=po.PFalse(
                ch='aom',
                dur=t_step,
                t0=self._relax
            )
        )
        count_trace_pb.append(
            p_obj=po.PFalse(
                ch='ctr_gate',
                dur=t_step,
            )
        )

        # Duration of full PB
        # [for gated_ctr.get_count_ar() timeout]
        self._pb_dur = count_trace_pb.dur

        # Default value
        count_trace_pb.dflt_dict = dict(
            aom=po.DFalse(),
            ctr_gate=po.DFalse()
        )

        return count_trace_pb

    def _make_iter(self):

        self._gated_ctr.start_counting()
        self._p_gen.start()

        # Read data from counter
        timeout = max(
            10 * self._pb_dur * self._hrdw_reps,
            1
        )
        flat_cnt_ar = self._gated_ctr.get_count_ar(timeout=timeout)

        # Sum counts from different hardware repetitions
        cnt_ar = np.reshape(
            a=flat_cnt_ar,
            newshape=(self._hrdw_reps, self._n_pts)
        )
        cnt_ar = np.sum(a=cnt_ar, axis=0)

        # Update data arrays
        self.raw_ar[self._soft_rep_idx][:] = cnt_ar
        self.sum_ar[:] = np.sum(
            a=self.raw_ar,
            axis=0
        )

        self._p_gen.stop()

    def _update_gui(self):

        if self._1d_trace is not None:
            self._1d_trace.set_data(
                x_ar=self.t_ar[:-1],
                y_ar=self.sum_ar
            )

        if self._2d_heat_map is not None:
            self._2d_heat_map.set_data(
                x_ar=self.t_ar[:-1],
                y_ar=np.arange(start=1, stop=self._soft_rep_idx + 1),
                z_ar=self.raw_ar[:self._soft_rep_idx, :-1]
            )

        if self._p_bar is not None:
            self._p_bar.set_value(
                value=100 * self._soft_rep_idx / self._soft_reps
            )

    def _hardware_off(self):
        self._p_gen.stop()
        self._gated_ctr.terminate_counting()



from pylabnet.utils.decorators.gui_decorators import handle_gui_errors
import pickle


class GUIHandler():
    """Generic GUI handler class providing an error-tolerant GUI client - script Interface.

    Member variables of this class keep track of connection parameters
    (is_running, is_connected, etc.), and global error handling is performed
    in the @handle_gui_errors decorator, which is applied to all calls
    from the GUIHandler to a GUI client.

    :gui_client: (object)
        GUI client to be called.
    :logger_client: (object)
        Logger client used for error logging in  @handle_gui_errors
        decorator.
    """

    def __init__(self,  gui_client=None, logger_client=None):

        self.is_running = False  # Flag which lets us know if WlmMonitor is running
        self.is_paused = False  # Flag which tells us we have simply paused WlmMonitor operation
        self.gui_connected = False  # Flag which alerts if a GUI client has been connected successfully
        self.gui_reconnect = False  # Flag which tells us to try reconnecting to the GUI client

        self.gui_client = gui_client
        self.logger_client = logger_client

    def assign_gui(self, gui_client):
        """Assigns a GUI client to the GUI handler

        :param client:
            (obj) instance of GUI client
        """

        self.gui = gui_client

    def pause(self):
        """Pauses the wavemeter monitor"""
        self.is_running = False
        self.is_paused = True

    def resume(self):
        """Resumes the wavemeter monitor when paused"""
        self.is_paused = False

    def reconnect_gui(self):
        """ Reconnects to the GUI

        Should be called if the GUI connection has been lost, once a new GUI client with the same access parameters has
        been reinstantiated
        """
        self.gui_reconnect = True

    # Functions called on gui client with corresponding error_handling decorator
    @handle_gui_errors
    def assign_plot(self, plot_widget, plot_label, legend_widget):
        return self.gui_client.assign_plot(
            plot_widget=plot_widget,
            plot_label=plot_label,
            legend_widget=legend_widget
        )

    @handle_gui_errors
    def clear_plot(self, plot_widget):
        return self.gui_client.clear_plot(
            plot_widget=plot_widget
        )

    @handle_gui_errors
    def assign_curve(self, plot_label, curve_label, error=False):
        return self.gui_client.assign_curve(
            plot_label=plot_label,
            curve_label=curve_label
        )

    @handle_gui_errors
    def remove_curve(self, plot_label, curve_label):
        return self.gui_client.remove_curve(
            plot_label=plot_label,
            curve_label=curve_label
        )

    @handle_gui_errors
    def assign_scalar(self, scalar_widget, scalar_label):
        self.gui_client.assign_scalar(
            scalar_widget=scalar_widget,
            scalar_label=scalar_label
        )

    @handle_gui_errors
    def assign_label(self, label_widget, label_label):
        return self.gui_client.assign_label(
            label_widget=label_widget,
            label_label=label_label
        )

    @handle_gui_errors
    def assign_event_button(self, event_widget, event_label):
        return self.gui_client.assign_event_button(
            event_widget=event_widget,
            event_label=event_label,
        )

    @handle_gui_errors
    def assign_event_button_event(self, event_label, function):
        return self.gui_client.assign_event_button_event(
            event_label=event_label,
            function=function,
        )

    @handle_gui_errors
    def set_curve_data(self, data, plot_label, curve_label, error=None):
        return self.gui_client.set_curve_data(
            data=data,
            plot_label=plot_label,
            curve_label=curve_label,
            error=error
        )

    @handle_gui_errors
    def set_scalar(self, value, scalar_label):
        return self.gui_client.set_scalar(
            value=value,
            scalar_label=scalar_label
        )

    @handle_gui_errors
    def get_scalar(self, scalar_label):
        return self.gui_client.get_scalar(scalar_label)

    @handle_gui_errors
    def activate_scalar(self, scalar_label):
        return self.gui_client.activate_scalar(scalar_label)

    @handle_gui_errors
    def deactivate_scalar(self, scalar_label):
        return self.gui_client.deactivate_scalar(scalar_label)

    @handle_gui_errors
    def set_label(self, text, label_label):
        return self.gui_client.set_label(
            text=text,
            label_label=label_label
        )

    @handle_gui_errors
    def was_button_pressed(self, event_label):
        return self.gui_client.was_button_pressed(event_label)



""" Implements connection and server launching of NI-daqMX card for wavemeter locking"""

import socket

from pylabnet.hardware.ni_daqs import nidaqmx_card
from pylabnet.network.client_server.nidaqmx_card import Service, Client
from pylabnet.network.core.generic_server import GenericServer

# Parameters
NI_DEVICE_NAME = 'cDAQ1Mod1'


def launch(**kwargs):
    """ Connects to NI-daqMX card and launches server

    :param kwargs: (dict) containing relevant kwargs
        :logger: instance of LogClient for logging purposes
        :port: (int) port number for the Cnt Monitor server
    """

    # Instantiate driver
    ni_daqmx_logger = kwargs['logger']
    try:
        ni_driver = nidaqmx_card.Driver(
            device_name=NI_DEVICE_NAME,
            logger=ni_daqmx_logger
        )
    except OSError:
        ni_daqmx_logger.error(f'Did not find NI daqMX name {NI_DEVICE_NAME}')
        raise

    # Instantiate server
    ni_daqmx_service = Service()
    ni_daqmx_service.assign_module(module=ni_driver)
    ni_daqmx_service.assign_logger(logger=ni_daqmx_logger)
    ni_daqmx_server = GenericServer(
        service=ni_daqmx_service,
        host=socket.gethostbyname(socket.gethostname()),
        port=kwargs['port']
    )

    ni_daqmx_server.start()



# -*- coding: utf-8 -*-


"""
This file contains wrappers for Jupyter Notebook Widgets used for controling the SC20 Shutter

"""

import ipywidgets as widgets


class ShutterToogle():
    """
    This Class implements a toggle switch to open and close the SC20 Shutter
    """

    def __init__(self, shutter_client):

        # retrieve shutter name
        self.name = shutter_client.get_name()

        # retrieve shutter client
        self.shutter_client = shutter_client

        # The toogle button
        self.toogle_raw = widgets.ToggleButtons(
            options=['Closed', 'Open'],
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
            tooltips=['Close Shutter', 'Open Shutter'],
            layout=widgets.Layout(
                width='50%',
                height='80px'
            )
        )

        # Add label to toogle
        self.toogle = widgets.HBox(
            [widgets.Label('Shutter {}:'.format(self.name)),
             self.toogle_raw]
            )


        self.output = widgets.Output()

        #Display Toogle
        display(self.toogle, self.output)

    def on_value_change(self, change):
        with self.output:
            if change['new'] == 'Open':
                return self.shutter_client.open()
            elif change['new'] == 'Closed':
                return self.shutter_client.close()



""" Script that launches and continuously runs a GUI window

Normally, this is meant to be invoked from within a Launcher object (see launcher.py).
However, you can also call this directly, with command-line arguments:
:arg --logport: port number of log server
:arg --guiport: (optional) port number to use for GUI server. Notes:
    (1) if not provided, the user will be prompted to enter a port number in the commandline
    (2) will raise a ConnectionRefusedError if port fails
:arg --ui: (optional) the name of the server module. Notes:
    (1) should be a valid .ui file (with .ui extension removed) within pylabnet/gui/pyqt/templates,
        otherwise, FileNotFound error will be raised
    (2) if not provided, _default_template will be used
"""

from PyQt5 import QtWidgets, QtCore
import sys
import socket
import numpy as np

from pylabnet.gui.pyqt.external_gui import Window
from pylabnet.network.client_server.external_gui import Service
from pylabnet.network.core.generic_server import GenericServer
from pylabnet.utils.logging.logger import LogClient
from pylabnet.utils.helper_methods import parse_args, show_console, hide_console

import sys
import socket

# Should help with scaling issues on monitors of differing resolution
if hasattr(QtCore.Qt, 'AA_EnableHighDpiScaling'):
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
if hasattr(QtCore.Qt, 'AA_UseHighDpiPixmaps'):
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


def main():

    # parse command line arguments
    args = parse_args()
    try:
        log_port = int(args['logport'])
    except IndexError:
        raise IndexError('logport not provided. Please provide command line arguments in the form\n"'
                         'python launch_gui.py --logip 000.000.000.000 --logport 1234 --guiport 5678 --ui uifilename')
    if 'logip' in args:
        log_ip = args['logip']
    else:
        log_ip = 'localhost'
    if 'ui' in args:
        gui_template = args['ui']
    else:
        show_console()
        gui_template = input('Please enter a GUI template to use: ')
        hide_console()
    if 'guiport' in args:
        gui_port = int(args['guiport'])
    else:
        show_console()
        gui_port = int(input('Please enter a GUI port value: '))
        hide_console()

    # Instantiate logger
    gui_logger = LogClient(
        host=log_ip,
        port=log_port,
        module_tag=gui_template+'_GUI',
        ui=gui_template,
        server_port=gui_port
    )

    # Retrieve debug flag.
    debug = int(args['debug'])

    # Halt execution and wait for debugger connection if debug flag is up.
    if debug:
        import ptvsd
        import os
        # 5678 is the default attach port in the VS Code debug configurations
        gui_logger.info(f"Waiting for debugger to attach to PID {os.getpid()} (pylabnet_gui)")
        ptvsd.enable_attach(address=('localhost', 5678))
        ptvsd.wait_for_attach()
        breakpoint()

    gui_logger.info('Logging for gui template: {}'.format(gui_template))

    # Create app and instantiate main window
    app = QtWidgets.QApplication(sys.argv)
    try:
        main_window = Window(app, gui_template=gui_template)
    except FileNotFoundError:
        gui_logger.warn('Could not find .ui file, '
                        'please check that it is in the pylabnet/gui/pyqt/gui_templates directory')
        raise

    # Instantiate GUI server
    gui_service = Service()
    gui_service.assign_module(module=main_window)
    gui_service.assign_logger(logger=gui_logger)

    # Make connection
    try:
        gui_server = GenericServer(
            service=gui_service,
            host=socket.gethostbyname(socket.gethostname()),
            port=gui_port
        )
    except ConnectionRefusedError:
        gui_logger.warn('Tried and failed to create GUI server with \nIP:{}\nPort:{}'.format(
            socket.gethostbyname(socket.gethostname()),
            gui_port
        ))
        raise
    gui_server.start()

    # Update GUI with server-specific details
    main_window.ip_label.setText('IP Address: {}'.format(
        socket.gethostbyname(socket.gethostname())
    ))
    main_window.port_label.setText('Port: {}'.format(gui_port))

    # Run the GUI until the stop button is clicked
    while not main_window.stop_button.isChecked():
        main_window.configure_widgets()
        main_window.update_widgets()
        main_window.force_update()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



""" pylabnet measurement and service classes for Swabian Instruments TimeTagger
which implements qudi's SlowCounter interface.

This file contains pylabnet wrapper and service classes to allow qudi to
access Swabian Instruments TT through pylabnet network as SlowCounter.

Steps:
- instantiate TimeTagger
- instantiate pylabnet-SlowCtrWrap (pass ref to TimeTagger as tagger)
- instantiate pylabnet-SlowCtrService and assign module to the created wrapper
- start pylabnet-server for SlowCtrService
- in qudi, instantiate SlowCtrClient as one of the hardware modules
"""

from pylabnet.network.core.service_base import ServiceBase
import TimeTagger as TT
import time
import copy
import pickle


class Wrap:
    """ Measurement instance which implements qudi's SlowCounter interface.
    """

    def __init__(self, tagger, channel_list, clock_frequency, buffer_size):

        # References to the device and to TT.Counter measurement
        self._tagger = tagger
        self._counter = None

        # Counter parameters
        self._channel_list = channel_list
        self._clock_frequency = clock_frequency
        self._buffer_size = buffer_size

        self._bin_width = 0
        self._bin_width_sec = 0

    def set_up_clock(self, clock_frequency=None, clock_channel=None):
        """
        Sets sample clock frequency for the Counter measurement.


        :param clock_frequency: (float) sample clock frequency. If not given,
                                configuration value is used
        :param clock_channel: ignored (internal timebase is used to generate
                              sample clock signal)

        :return: (int) operation status code: 0 - OK
                                             -1 - Error
        """

        # Use config value, if no clock_frequency is specified
        if clock_frequency is None:
            clock_frequency = self._clock_frequency

        # Calculate final bin width
        bin_width = int(1e12/clock_frequency)  # in picoseconds, for device
        bin_width_sec = bin_width * 1e-12      # is seconds, for software timing

        # Set new values param to internal variables
        self._bin_width = bin_width
        self._bin_width_sec = bin_width_sec

        return 0

    def set_up_counter(self,
                       counter_channels=None,
                       sources=None,
                       clock_channel=None,
                       counter_buffer=None):
        """
        Configures the actual counter with a given clock.


         (list of int) [optional] list of channels
                     to count clicks on. If not given, config value is used.

        :param counter_buffer: (int) [optional] size of the memory buffer.
                               If not given, config value is used.

        :param counter_channels: ignored
            This argument should not be used. Counter GUI initializes set of plot curves
            self.curves during its on_activate() method. It basically calls
            counter_hardware.get_counter_channels() and uses this list to init self.curves
            Only after that user can click "Start" button, which will call set_up_counter().
            And since GUI already has inited set of curves, set of channels must not be
            modified here! It will case GUI to fail.

        :param sources: ignored
        :param clock_channel: ignored

        :return: (int) operation status code: 0 - OK
                                             -1 - Error
        """

        # Set counter channels
        if counter_channels is not None:
            channel_list = counter_channels
        else:
            channel_list = self._channel_list
        # apply counter channel change
        self.set_counter_channels(channel_list=channel_list)

        # Set buffer size
        if counter_buffer is not None:
            buffer_size = counter_buffer
        else:
            buffer_size = self._buffer_size
        # sanity check:
        if not isinstance(buffer_size, int) or buffer_size<=0:
            # self.log.error('set_up_counter(): invalid parameter value counter_buffer = {}.'
            #                'This parameter must be a positive integer.'
            #                ''.format(buffer_size))
            return -1
        # apply buffer size change
        self._buffer_size = buffer_size

        # Create instance of Counter measurement
        try:
            self._counter = TT.Counter(
                tagger=self._tagger,
                channels=self._channel_list,
                binwidth=self._bin_width,
                n_values=self._buffer_size
            )
        # handle initialization error (TT functions always produce NotImplementedError)
        except NotImplementedError:
            self._counter = None
            # self.log.error('set_up_counter(): failed to instantiate TT.Counter measurement')
            return -1

        # Start Counter
        # (TT.Counter measurement starts running immediately after instantiation,
        # so it is necessary to erase all counts collected since instantiation)
        self._counter.stop()
        self._counter.clear()
        self._counter.start()

        return 0

    def close_clock(self):
        """
        Closes the clock.

        :return: (int) error code: 0 - OK
                                  -1 - Error
        """

        # self._bin_width = 0
        # self._bin_width_sec = 0

        return 0

    def close_counter(self):
        """
        Closes the counter and cleans up afterwards.

        :return: (int) error code: 0 - OK
                                  -1 - Error
        """

        # Try stopping and clearing TT.Counter measurement
        try:
            self._counter.stop()
            self._counter.clear()
        # Handle the case of exception in TT function call (NotImplementedError)
        # and the case of self._ctr = None (AttributeError)
        except (NotImplementedError, AttributeError):
            pass

        # Remove reference to the counter
        # self._ctr = None

        # Clear counter parameters
        # self._buffer_size = []

        # Do not clear channel list:
        # Counter GUI inits its list of curves self.curves
        # by calling counter_hardware.get_counter_channels() before
        # calling counter_hardware.set_up_counter()
        # If one clears _channel_list here, GUI will fail at the next
        # "Start" button click after reloading.
        #
        # self._channel_list = []

        return 0

    def get_counter(self, samples=1):
        """
        Returns the current counts per second of the counter.

        :param samples: (int) [optional] number of samples to read in one go
                        (default is one sample)

        :return: numpy.array((samples, uint32), dtype=np.uint32)
        array of count rate [counts/second] arrays of length samples for each click channel
        Empty array [] is returned in the case of error.
        """

        # Sanity check: samples has valid value
        if samples != 1:
            if not isinstance(samples, int) or samples <= 0:
                # self.log.error('get_counter(): invalid argument samples={0}. This argument must be a positive integer'
                #                ''.format(samples))
                return []

        # MORE SOPHISTICATED VERSION
        # (WORKS TOO SLOWLY: PROBABLY BECAUSE OF SLOW INTEGER DIVISION OF LARGE INTEGERS)
        #
        # start_time = time.time()
        # while time.time() - start_time < self._timeout:
        #     new_complete_bins = self._ctr.getCaptureDuration() // self._bin_width - self._last_read_bin
        #
        #     self._overflow = new_complete_bins
        #     # self.log.error('new_complete_bins = {}'.format(new_complete_bins))
        #
        #     if new_complete_bins < samples:
        #         time.sleep(self._bin_width_sec/2)
        #         continue
        #     elif new_complete_bins == samples:
        #         self._last_read_bin += new_complete_bins
        #         break
        #     else:
        #         # self.log.warn('Counter is overflowing. \n'
        #         #               'Software pulls data in too slowly and counter bins are too short, '
        #         #               'such that some bins are lost. \n'
        #         #               'Try reducing sampling rate or increasing oversampling')
        #         self._last_read_bin += new_complete_bins
        #         break

        # Wait for specified number of samples (samples parameter) to be accumulated
        #
        # This approach is very naive and is more or less accurate for
        # clock frequency below 50 Hz.
        #
        # For higher frequencies, the actual time sampling interval is determined
        # by software delays (about 1 ms). Counter measurement overflows
        # (most of the samples are over-written before software reads them in)
        # but does not fail. The only problem here is that time axis on the count-trace
        # graph is no longer accurate:
        # the difference between consecutive tick labels is much smaller than the actual
        # time interval between measured samples (about 1 ms)
        time.sleep(samples * self._bin_width_sec)

        # read-in most recent 'samples' samples
        try:
            count_array = self._counter.getData()[:, -samples:]
        except NotImplementedError:
            # self.log.error('get_counter() reading operation failed')
            return []
        except AttributeError:
            # self.log.error('get_counter(): counter was not initialized')
            return []

        # Calculate count rate [count/sec]
        count_rate_array = count_array / self._bin_width_sec

        return count_rate_array

    def get_counter_channels(self):
        """
        Returns the list of click channel numbers.

        :return: (list of int) list of click channel numbers
        """

        return copy.deepcopy(self._channel_list)

    def set_counter_channels(self, channel_list=None):
        """
        Set click channel list.

        Notice that this method only modifies internal variable _channel_list.
        To apply the change to the counter, one has to call set_up_counter() again.


        :param channel_list: (list of int) list of channels to count clicks on

        :return: (list of int) actual list of click channels
        """

        if channel_list is None:
            return self.get_counter_channels()

        # Sanity check:
        all_channels = self._get_all_channels()
        if not set(channel_list).issubset(set(all_channels)):
            # self.log.error('set_counter_channels(): requested list of channels is invalid: '
            #                'some channels are not present on the device.'
            #                'requested list: {0} \n'
            #                'available channels: {1}'
            #                ''.format(channel_list, all_channels))
            return self.get_counter_channels()

        # Apply changes to internal variable self._channel_list
        self._channel_list = channel_list
        # Sort channel numbers, such that channel order does not depend
        # on order of numbers in the config file
        self._channel_list.sort()

        return self.get_counter_channels()

    def _get_all_channels(self):
        """
        Return list of all channels available on the device.

        Positive/negative values correspond to rising/falling edge detection.
        For example:
            1 means 'rising edge on connector 1'
            -1 means 'falling edge on connector 1


        :return: (list of int) list of all available channel numbers,
                               including edge sign.
        """

        try:
            available_channel_tuple = list(
                self._tagger.getChannelList(TT.TT_CHANNEL_RISING_AND_FALLING_EDGES)
            )
        # handle exception in the call (TT functions normally produce NotImplementedError)
        except NotImplementedError:
            # self.log.error('_get_all_channels(): communication with the device failed')
            return []
        # handle the case of self._tagger = None
        except AttributeError:
            # self.log.error('_get_all_channels(): _tagger is None. Initialize device first')
            return []

        return list(available_channel_tuple)


class Service(ServiceBase):

    def exposed_set_up_clock(self, clock_frequency=None, clock_channel=None):
        """
        Sets sample clock frequency for the Counter measurement.


        :param clock_frequency: (float) sample clock frequency. If not given,
                                configuration value is used
        :param clock_channel: ignored (internal timebase is used to generate
                              sample clock signal)

        :return: (int) operation status code: 0 - OK
                                             -1 - Error
        """

        return self._module.set_up_clock(
            clock_frequency=clock_frequency,
            clock_channel=clock_channel
        )

    def exposed_set_up_counter(self, counter_channels=None, sources=None, clock_channel=None, counter_buffer=None):
        """
        Configures the actual counter with a given clock.


         (list of int) [optional] list of channels
                     to count clicks on. If not given, config value is used.

        :param counter_buffer: (int) [optional] size of the memory buffer.
                               If not given, config value is used.

        :param counter_channels: ignored
            This argument should not be used. Counter GUI initializes set of plot curves
            self.curves during its on_activate() method. It basically calls
            counter_hardware.get_counter_channels() and uses this list to init self.curves
            Only after that user can click "Start" button, which will call set_up_counter().
            And since GUI already has inited set of curves, set of channels must not be
            modified here! It will case GUI to fail.

        :param sources: ignored
        :param clock_channel: ignored

        :return: (int) operation status code: 0 - OK
                                             -1 - Error
        """

        return self._module.set_up_counter(
            counter_channels=counter_channels,
            sources=sources,
            clock_channel=clock_channel,
            counter_buffer=counter_buffer
        )

    def exposed_close_clock(self):
        """
        Closes the clock.

        :return: (int) error code: 0 - OK
                                  -1 - Error
        """

        return self._module.close_clock()

    def exposed_close_counter(self):
        """
        Closes the counter and cleans up afterwards.

        :return: (int) error code: 0 - OK
                                  -1 - Error
        """

        return self._module.close_ctr()

    def exposed_get_counter(self, samples=1):
        """
        Returns the current counts per second of the counter.

        :param samples: (int) [optional] number of samples to read in one go
                        (default is one sample)

        :return: numpy.array((samples, uint32), dtype=np.uint32)
        array of count rate [counts/second] arrays of length samples for each click channel
        Empty array [] is returned in the case of error.
        """

        res = self._module.get_counter(samples=samples)
        return pickle.dumps(res)

    def exposed_get_counter_channels(self):
        """
        Returns the list of click channel numbers.

        :return: (list of int) list of click channel numbers
        """

        res = self._module.get_counter_channels()
        return pickle.dumps(res)



import numpy as np


n1 = 5
n2 = 3
n3 = 2

ar1 = np.linspace(start=0, stop=n1-1, num=n1)
ar2 = np.linspace(start=0, stop=n2-1, num=n2)
ar3 = np.linspace(start=0, stop=n3-1, num=n3)

idx1 = 0
idx2 = 0
idx3 = 0


if idx1 < n1 and idx2 < n2 and idx3 < n3:

    stop_req = False

    while True:
        # Perform operation
        print('p1 = {} ================'.format(ar1[idx1]))

        while True:
            # Perform operation
            print('p2 = {} -------'.format(ar2[idx2]))

            while True:
                # Perform operation
                print('p3 = {}'.format(ar3[idx3]))

                if stop_req:
                    print('Loop 3 stopped')
                    break
                else:
                    idx3 += 1
                    if idx3 == n3:
                        idx3 = 0
                        break

            if stop_req:
                print('Loop 2 stopped')
                break
            else:
                idx2 += 1
                if idx2 == n2:
                    idx2 = 0
                    break

        if stop_req:
            print('Loop 1 stopped')
            break
        else:
            idx1 += 1
            if idx1 == n1:
                idx1 = 0
                break

    if stop_req:
        print('Stop hardware')


# Level template
# while True:
#     # Operation
#     # ========
#
#     # Index incrementation
#     if stop_req:
#         break
#     else:
#         idx += 1
#         if idx == n:
#             idx = 0
#             break



# -*- coding: utf-8 -*-


"""This file contains the pylabnet Hardware module class for Thorlabs FW102c Filterwheel.
Is is essentially a wrapper for the class FW102C in fw102c.py writen by Gilles Simond
"""

from pylabnet.hardware.filterwheel.fw102c import FW102C
from pylabnet.utils.logging.logger import LogHandler
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase


class FW102CFilterWheel:

    def __init__(self, port_name, device_name, filters, logger=None):
        """Instantiate Hardware class for FW102c Filterwheel by instantiating a FW102C class

        :device_name: Readable name of device, e.g. 'Collection Filters'
        :port_name: Port name over which Filter wheel is connect via USB
        :filters: A dictionary where the keys are the numbered filter positions
                  and the values are strings describing the filter, e.g. '4 ND'
        """

        # Instantiate log
        self.log = LogHandler(logger=logger)

        # Retrieve name and filter options
        self.device_name = device_name
        self.filters = filters

        # Instantiate FW102C
        self.filter_wheel = FW102C(port=port_name, logger=self.log)

        if not self.filter_wheel.isOpen:
            self.log.error("Filterwheel {} connection failed".format(self.device_name))
        else:
            self.log.info("Filterwheel {} connection successfully".format(self.device_name))

    def get_name(self):
        return self.device_name

    def get_pos(self):
        """Returns current position of filterwheel
        """

        return self.filter_wheel.query('pos?')

    def change_filter(self, new_pos, protect_shutter_client):
        """ Update filter wheel position

        :new_pos: Target filter wheel position 1-6 or 1-12
        :protect_shutter_client: (optional) SC20Shutter instance.
            If provided, shutter will be closed during a filter change
        """

        # Close protection shutter if open
        if protect_shutter_client is not None:
            # Check if shutter is initially open
            shutter_open = protect_shutter_client.get_is_open()

            if shutter_open:
                protect_shutter_client.close()

        # Update Position
        self.filter_wheel.command('pos={}'.format(new_pos))

        # Check if update was successful
        if int(self.get_pos()) == int(new_pos):
            self.log.info(
                "Filterwheel {device_name} changed to position {position} ({filter})"
                "".format(
                    device_name=self.device_name,
                    position=new_pos,
                    filter=self.filters.get('{}'.format(new_pos)))
            )

            # Open protection shutter if shutter was originally open
            if protect_shutter_client is not None and shutter_open:
                protect_shutter_client.open()
        else:
            self.log.error(
                "Filterwheel {device_name}: changing to position failed"
                "".format(device_name=self.device_name)
            )

    # returns filter dictionary
    def get_filter_dict(self):
        return self.filters


class Service(ServiceBase):

    def exposed_change_filter(self, new_pos, protect_shutter_client):
        return self._module.change_filter(new_pos, protect_shutter_client)

    def exposed_get_pos(self):
        return self._module.get_pos()

    def exposed_get_filter_dict(self):
        return self._module.get_filter_dict()

    def exposed_get_name(self):
        return self._module.get_name()


class Client(ClientBase):
    def change_filter(self, new_pos, protect_shutter_client=None):
        return self._service.exposed_change_filter(new_pos, protect_shutter_client)

    def get_pos(self):
        return self._service.exposed_get_pos()

    def get_filter_dict(self):
        return self._service.exposed_get_filter_dict()

    def get_name(self):
        return self._service.exposed_get_name()



from pylabnet.utils.logging.logger import LogHandler
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase
from pylabnet.network.core.generic_server import GenericServer


class Driver():

    def __init__(self, name, logger,  hardware_module,  **kwargs):
        '''High level staticline class.

        This class is used in conjunction with hardware modules to send out digital
        signals ('voltage low' and 'voltage high'). This top level class is hardware agnostic.
        With the use of a StaticLineHardwareHandler, this class will be associated with the necessary
        setup functions and output functions of a hardware module.

        :name: A easily recognizable name for this staticline, ideally referring to the device being controlled
            by it, e.g. 'Shutter 1'.
        :logger: An instance of a LogClient.
        :hardware_module: An instance of a hardware module. Based on this, the corresponsind setup function is automatically
            called.
        '''

        self.name = name

        # Instantiate log
        self.log = LogHandler(logger=logger)

        # Instantiate Hardware_handler
        self.hardware_handler = StaticLineHardwareHandler(
            hardware_module,
            self.log, name,
            **kwargs
        )

    def up(self):
        '''Set output to high.'''
        self.hardware_handler.up()
        self.log.info(
            f"Staticline {self.name} set to high."
        )

    def down(self):
        '''Set output to low.'''
        self.hardware_handler.down()
        self.log.info(
            f"Staticline {self.name} set to low."
        )

    def get_name(self):
        return self.name


class StaticLineHardwareHandler():

    def _HDAWG_toggle(self, newval):
        ''' Set DIO_bit to high or low

        :newval: Either 0 or 1 indicating the new output state.
        '''

        # Get current DIO output integer.
        current_output = self.hardware_module.geti('dios/0/output')

        if newval == 0:
            # E.g., for DIO-bit 3: 1111 .... 1110111
            DIO_bit_bitshifted = ~(0b1 << self.DIO_bit)

            # Binary AND generates new output.
            new_output = current_output & DIO_bit_bitshifted

        elif newval == 1:
            # E.g., for DIO-bit 3: 0000 ... 0001000
            DIO_bit_bitshifted = (0b1 << self.DIO_bit)

            # Binary OR generates new output.
            new_output = current_output | DIO_bit_bitshifted

        self.hardware_module.seti('dios/0/output', new_output)

    def _setup_HDWAGDriver(self, **kwargs):
        ''' Setup a ZI HDAWG driver module to be used as a staticline toggle.

        :DIO_bit: Which bit to toggle, in decimal notation.
        '''

        # Retrieve arguments from keyword argument dictionary.
        DIO_bit = kwargs['DIO_bit']

        # Drive 8-bit bus containing DIO_bit to be toggled.
        # Note that the buses are enabled by using the decimal equivalent
        # of a binary number indicating which bus is driven:
        # 1101 = 11 corresponds to driving bus 1, 2, and 4.

        if DIO_bit in range(8):
            toggle_bit = 1  # 1000
        elif DIO_bit in range(8, 16):
            toggle_bit = 2  # 0100
        elif DIO_bit in range(16, 24):
            toggle_bit = 4  # 0010
        elif DIO_bit in range(24, 32):
            toggle_bit = 8  # 0001
        else:
            self.log.error(f"DIO_bit {DIO_bit} invalid, must be in range 0-31.")

        self.DIO_bit = DIO_bit
        self.log.info(f"DIO_bit {DIO_bit} successfully assigned to staticline {self.name}.")

        # Read in current configuration of DIO-bus.
        current_config = self.hardware_module.geti('dios/0/drive')

        # Set new configuration by using the bitwise OR.
        new_config = current_config | toggle_bit
        self.hardware_module.seti('dios/0/drive', new_config)

        # Register up/down function.
        self.up = lambda: self._HDAWG_toggle(1)
        self.down = lambda: self._HDAWG_toggle(0)

    def _setup_NiDaqMxDriver(self, **kwargs):

        # Retrieve arguments from keyword argument dictionary,
        # if not found apply default value.
        try:
            down_voltage = kwargs['down_voltage']
        except KeyError:
            down_voltage = 0

        try:
            up_voltage = kwargs['up_voltage']
        except KeyError:
            up_voltage = 3.3

        # Check if voltages are in bound.
        if not -10 <= down_voltage <= 10:
            self.log.error(f'Down voltage of {down_voltage} V is invalid, must be between -10 V and 10 V.')
        if not -10 <= up_voltage <= 10:
            self.log.error(f'Up voltage of {up_voltage} V is invalid, must be between -10 V and 10 V.')

        ao_output = kwargs['ao_output']

        # Register up/down function.
        ao_output = kwargs['ao_output']
        self.up = lambda: self.hardware_module.set_ao_voltage(ao_output, up_voltage)
        self.down = lambda: self.hardware_module.set_ao_voltage(ao_output, down_voltage)

        # Set voltage to down.
        self.down()

        # Log successfull setup.
        self.log.info(f"NiDaq {self.hardware_module.dev} output {ao_output} successfully assigned to staticline {self.name}.")

    def __init__(self, hardware_module, loghandler, name, **kwargs):
        '''Handler connecting hardware class to StaticLine instance

        Main task of this instance is to define the device-specific function
        which should correspond to setting the staticline to high or low, and
        to set up the hardware accordingly.

        :hardware_module: Hardware module to be used to toggle the staticline.
        :loghandler: Instance of loghandler.
        :name: Name of StaticLine instance.
        :**kwargs: Additional keyword arguments which depending on the hardware module
            contain values needed to setup the hardware as a staticline.
        '''
        self.hardware_module = hardware_module
        self.name = name
        self.log = loghandler

        # Read string of module name (e.g. 'HDAWGDriver').
        self.hardware_module_name = str(hardware_module.__class__).split('.')[-2]

        # Dictionary listing all hardware modules which can address
        # staticlines and their corresponding setup functions.
        registered_staticline_modules = {
            'zi_hdawg':  self._setup_HDWAGDriver,
            'nidaqmx_card': self._setup_NiDaqMxDriver
        }

        # Check if hardware module is registered.
        if self.hardware_module_name not in registered_staticline_modules.keys():
            valid_modules = list(registered_staticline_modules.keys())
            error_msg = f"Setup of staticline using module {self.hardware_module_name} failed. Compatible modules are {valid_modules}."
            self.log.error(error_msg)

        # Depending on which module is used,
        # automatically call the hardware-specific setup function.
        for module_name, setup_function in registered_staticline_modules.items():

            if self.hardware_module_name == module_name:

                # Call hardware specific setup function.
                setup_function(**kwargs)

                self.log.info(
                    f"Setup of staticline {name} using module {module_name} successful."
                )


""" Driver for Zurich Instruments HDAWG """

import zhinst.ziPython as zi


class Driver:

    def __init__(self, dev='DEV8040'):

        # Start ZI DAQ server
        self._discovery = zi.ziDiscovery()
        self.device_id = self._discovery.find(dev)
        device_props = self._discovery.get(self.device_id)



from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase
from pylabnet.hardware.interface.wavemeter import WavemeterInterface


class Service(ServiceBase):

    def exposed_get_wavelength(self, channel, units):
        return self._module.get_wavelength(
            channel=channel,
            units=units
        )


class Client(ClientBase, WavemeterInterface):

    def get_wavelength(self, channel=1, units="Frequency(THz)"):
        return self._service.exposed_get_wavelength(channel, units)



import plotly.offline as pltly
import plotly.graph_objs as graph_objs
pltly.init_notebook_mode()


def iplot(pb_obj, use_gl=False):

    # Do nothing for empty PulseBlock
    if len(pb_obj.dflt_dict.keys()) == 0 and len(pb_obj.p_dict.keys()) == 0:
        return None

    # Iterate through p_dict.keys() and dflt_dict.keys()
    # and create a trace for each channel
    #  - create sorted list of channels
    d_ch_set = set(pb_obj.dflt_dict.keys())
    p_ch_set = set(pb_obj.p_dict.keys())
    ch_list = list(d_ch_set | p_ch_set)
    ch_list.sort()

    # - iterate trough ch_list
    trace_list = []
    for ch_index, ch in enumerate(ch_list):

        #
        # Build x_ar, y_ar, text_ar
        #

        # initial zero-point - default pulse object printout
        x_ar = [0]
        y_ar = [ch_index]
        if ch in pb_obj.dflt_dict.keys():
            text_ar = [
                '{}'.format(
                    str(pb_obj.dflt_dict[ch])
                )
            ]
        else:
            text_ar = ['']

        # Iterate through pulse list and create a rectangular
        # arc for each pulse. The mid-point on the upper segment
        # contains printout of the pulse object
        if ch in pb_obj.p_dict.keys():
            for p_item in pb_obj.p_dict[ch]:
                # edges of the pulse
                t1 = p_item.t0
                t2 = p_item.t0 + p_item.dur

                # left vertical line
                if t1 == 0:
                    # If pulse starts at the origin,
                    # do not overwrite (x=0, y=ch_index) point
                    # which contains dflt_dict[ch] printout
                    x_ar.append(t1)
                    y_ar.append(ch_index + 0.8)
                else:
                    x_ar.extend([t1, t1])
                    y_ar.extend([ch_index, ch_index + 0.8])

                # mid-point, which will contain printout
                x_ar.append((t1 + t2) / 2)
                y_ar.append(ch_index + 0.8)

                # right vertical line
                x_ar.extend([t2, t2])
                y_ar.extend([ch_index + 0.8, ch_index])

                # set mid-point text to object printout
                if t1 == 0:
                    # If pulse starts at the origin,
                    # do not overwrite (x=0, y=ch_index) point
                    # which contains dflt_dict[ch] printout
                    text_ar.extend(
                        [
                            '{:.2e}'.format(t1),
                            '{}'.format(str(p_item)),
                            '{:.2e}'.format(t2),
                            '{:.2e}'.format(t2)
                        ]
                    )
                else:
                    text_ar.extend(
                        [
                            '{:.2e}'.format(t1),
                            '{:.2e}'.format(t1),
                            '{}'.format(str(p_item)),
                            '{:.2e}'.format(t2),
                            '{:.2e}'.format(t2)
                        ]
                    )

        # final zero-point
        x_ar.append(pb_obj.dur)
        y_ar.append(ch_index)
        text_ar.append('{:.2e}'.format(pb_obj.dur))

        #
        # END: Build x_ar, y_ar, text_ar
        #

        # Create trace for 'ch'
        if use_gl:
            # For a large PulseBlock (above 10k pulses)- use WebGL module
            # Notice that there is a limit on number of WebGL objects in a
            # notebook (about 7-9), so use this mode for large blocks only
            trace = graph_objs.Scattergl(
                x=x_ar,
                y=y_ar,
                text=text_ar,
                hoverinfo='text',
                mode="lines",
                name=ch
            )
        else:
            # Standard Scatter for not too large PulseBlock objects
            trace = graph_objs.Scatter(
                x=x_ar,
                y=y_ar,
                text=text_ar,
                hoverinfo='text',
                mode="lines",
                name=ch
            )
        trace_list.append(trace)

    # Create figure
    fig = graph_objs.Figure(data=trace_list)

    fig['layout'].update(title='{}'.format(pb_obj.name))
    fig['layout'].update(height=100 + 100 * len(ch_list), width=800)
    fig['layout']['yaxis'].update(showticklabels=False)
    fig['layout']['legend'].update(traceorder='reversed')
    fig['layout'].update(hovermode='closest')

    # Display figure
    config_dict = {
        'showLink': False,
        'displaylogo': False
    }
    pltly.iplot(figure_or_data=fig, config=config_dict)
